# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rp2.plugin.country.us import US
from rp2.rp2_decimal import ZERO, RP2Decimal

from dali.configuration import Keyword
from dali.in_transaction import InTransaction
from dali.out_transaction import OutTransaction
from dali.plugin.input.csv.binance_com_supplemental import InputPlugin


class TestBinanceCsv:
    def test_autoinvest(self) -> None:
        plugin = InputPlugin(
            account_holder="tester",
            autoinvest_csv_file="input/test_binance_autoinvest.csv",
            betheth_csv_file=None,
            native_fiat="USD",
        )

        result = plugin.load(US())

        # 1 ETH in transaction for 0.01 ETH +
        # 1 USDT out transaction for 0.01 ETH +
        # 1 BTC in transaction for 0.002 BTC +
        # 1 BTC out transaction for 0.002 BTC = 4
        assert len(result) == 4

        eth_transaction_in: InTransaction = result[0]  # type: ignore
        eth_transaction_out: OutTransaction = result[1]  # type: ignore
        btc_transaction_in: InTransaction = result[2]  # type: ignore
        btc_transaction_out: OutTransaction = result[3]  # type: ignore

        # Buy ETH autoinvest
        assert eth_transaction_in.asset == "ETH"
        assert eth_transaction_in.timestamp == "2022-05-01 14:00:00+0000"
        assert eth_transaction_in.transaction_type == Keyword.BUY.value.capitalize()
        assert eth_transaction_in.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(eth_transaction_in.crypto_in) == RP2Decimal("0.01")
        assert eth_transaction_in.crypto_fee is None
        assert eth_transaction_in.fiat_in_no_fee is None
        assert eth_transaction_in.fiat_in_with_fee is None
        assert eth_transaction_in.fiat_fee is None

        assert eth_transaction_out.asset == "USDT"
        assert eth_transaction_out.timestamp == "2022-05-01 14:00:00+0000"
        assert eth_transaction_out.transaction_type == Keyword.SELL.value.capitalize()
        assert eth_transaction_out.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(eth_transaction_out.crypto_out_no_fee) == RP2Decimal("10.00")
        assert RP2Decimal(eth_transaction_out.crypto_fee) == RP2Decimal("0.0025")
        assert RP2Decimal(str(eth_transaction_out.crypto_out_with_fee)) == RP2Decimal("10.0025")
        assert eth_transaction_out.fiat_out_no_fee is None
        assert eth_transaction_out.fiat_fee is None

        # Buy BTC autoinvest
        assert btc_transaction_in.asset == "BTC"
        assert btc_transaction_in.timestamp == "2022-05-01 14:00:00+0000"
        assert btc_transaction_in.transaction_type == Keyword.BUY.value.capitalize()
        assert btc_transaction_in.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(btc_transaction_in.crypto_in) == RP2Decimal("0.002")
        assert btc_transaction_in.crypto_fee is None
        assert btc_transaction_in.fiat_in_no_fee is None
        assert btc_transaction_in.fiat_in_with_fee is None
        assert btc_transaction_in.fiat_fee is None

        assert btc_transaction_out.asset == "USDT"
        assert btc_transaction_out.timestamp == "2022-05-01 14:00:00+0000"
        assert btc_transaction_out.transaction_type == Keyword.SELL.value.capitalize()
        assert btc_transaction_out.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(btc_transaction_out.crypto_out_no_fee) == RP2Decimal("20.00")
        assert RP2Decimal(btc_transaction_out.crypto_fee) == RP2Decimal("0.005")
        assert RP2Decimal(str(btc_transaction_out.crypto_out_with_fee)) == RP2Decimal("20.005")
        assert btc_transaction_out.fiat_out_no_fee is None
        assert btc_transaction_out.fiat_fee is None

    def test_betheth(self) -> None:
        plugin = InputPlugin(
            account_holder="tester",
            autoinvest_csv_file=None,
            betheth_csv_file="input/test_binance_betheth.csv",
            native_fiat="USD",
        )

        result = plugin.load(US())

        # 1 ETH out transaction for 0.1 ETH +
        # 1 BETH in transaction for 0.1 BETH = 2
        assert len(result) == 2

        betheth_transaction_in: InTransaction = result[0]  # type: ignore
        betheth_transaction_out: OutTransaction = result[1]  # type: ignore

        assert betheth_transaction_in.asset == "BETH"
        assert betheth_transaction_in.timestamp == "2021-03-01 12:00:00+0000"
        assert betheth_transaction_in.transaction_type == Keyword.BUY.value.capitalize()
        assert betheth_transaction_in.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(betheth_transaction_in.crypto_in) == RP2Decimal("0.1")
        assert betheth_transaction_in.crypto_fee is None
        assert betheth_transaction_in.fiat_in_no_fee is None
        assert betheth_transaction_in.fiat_in_with_fee is None
        assert betheth_transaction_in.fiat_fee is None

        assert betheth_transaction_out.asset == "ETH"
        assert betheth_transaction_out.timestamp == "2021-03-01 12:00:00+0000"
        assert betheth_transaction_out.transaction_type == Keyword.SELL.value.capitalize()
        assert betheth_transaction_out.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(betheth_transaction_out.crypto_out_no_fee) == RP2Decimal("0.1")
        assert RP2Decimal(betheth_transaction_out.crypto_fee) == ZERO
        assert RP2Decimal(str(betheth_transaction_out.crypto_out_with_fee)) == RP2Decimal("0.1")
        assert betheth_transaction_out.fiat_out_no_fee is None
        assert betheth_transaction_out.fiat_fee is None



# Copyright 2022 Neal Chambers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from datetime import datetime, timedelta
from inspect import Signature, signature
from time import sleep, time
from typing import Any, Dict, List, NamedTuple, Optional, Union

from ccxt import (
    DDoSProtection,
    Exchange,
    ExchangeError,
    ExchangeNotAvailable,
    NetworkError,
    RequestTimeout,
    binance,
    gateio,
    huobi,
    kraken,
)
from rp2.logger import create_logger
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2RuntimeError

from dali.abstract_pair_converter_plugin import (
    AbstractPairConverterPlugin,
    AssetPairAndTimestamp,
)
from dali.configuration import Keyword
from dali.historical_bar import HistoricalBar
from dali.plugin.pair_converter.csv.kraken import Kraken as KrakenCsvPricing

# Native format keywords
_ID: str = "id"
_BASE: str = "base"
_QUOTE: str = "quote"
_SYMBOL: str = "symbol"
_TYPE: str = "type"

# Time in ms
_MINUTE: str = "1m"
_FIVE_MINUTE: str = "5m"
_FIFTEEN_MINUTE: str = "15m"
_ONE_HOUR: str = "1h"
_FOUR_HOUR: str = "4h"
_ONE_DAY: str = "1d"
_TIME_GRANULARITY: List[str] = [_MINUTE, _FIVE_MINUTE, _FIFTEEN_MINUTE, _ONE_HOUR, _FOUR_HOUR, _ONE_DAY]
_TIME_GRANULARITY_IN_SECONDS: List[int] = [60, 300, 900, 3600, 14400, 86400]

# Currently supported exchanges
_BINANCE: str = "Binance.com"
_GATE: str = "Gate"
_HUOBI: str = "Huobi"
_KRAKEN: str = "Kraken"
_FIAT_EXCHANGE: str = "Exchangerate.host"
_DEFAULT_EXCHANGE: str = _KRAKEN
_EXCHANGE_DICT: Dict[str, Any] = {_BINANCE: binance, _GATE: gateio, _HUOBI: huobi, _KRAKEN: kraken}

# Delay in fractional seconds before making a request to avoid too many request errors
# Kraken states it has a limit of 1 call per second, but this doesn't seem to be correct.
# It appears Kraken public API is limited to around 12 calls per minute.
# There also appears to be a limit of how many calls per 2 hour time period.
# Being authenticated lowers this limit.
_REQUEST_DELAYDICT: Dict[str, float] = {_KRAKEN: 5.1}

# CSV Pricing classes
_CSV_PRICING_DICT: Dict[str, Any] = {_KRAKEN: KrakenCsvPricing}

# Alternative Markets and exchanges for stablecoins or untradeable assets
_ALT_MARKET_EXCHANGES_DICT: Dict[str, str] = {
    "ATDUSDT": _GATE,
    "BETHETH": _BINANCE,
    "BNBUSDT": _BINANCE,
    "BSVUSDT": _GATE,
    "BOBAUSD": _GATE,
    "BUSDUSDT": _BINANCE,
    "EDGUSDT": _GATE,
    "ETHWUSD": _KRAKEN,
    "NEXOUSDT": _BINANCE,
    "SGBUSD": _KRAKEN,
    "SOLOUSDT": _HUOBI,
    "USDTUSD": _KRAKEN,
}

_ALT_MARKET_BY_BASE_DICT: Dict[str, str] = {
    "ATD": "USDT",
    "BETH": "ETH",
    "BNB": "USDT",
    "BOBA": "USD",
    "BSV": "USDT",
    "BUSD": "USDT",
    "EDG": "USDT",
    "ETHW": "USD",
    "NEXO": "USDT",
    "SGB": "USD",
    "SOLO": "USDT",
    "USDT": "USD",
}

# Priority for quote asset. If asset is not listed it will be filtered out.
# In principle this should be fiat in order of trade volume and then stable coins in order of trade volume
_QUOTE_PRIORITY: Dict[str, None] = {
    "USD": None,
    "JPY": None,
    "KRW": None,
    "EUR": None,
    "GBP": None,
    "AUD": None,
    "USDT": None,
    "USDC": None,
    "BUSD": None,
    "TUSD": None,
    "OUSD": None,
}

# Time constants
_MS_IN_SECOND: int = 1000

# Cache
_CACHE_INTERVAL: int = 200

# CSV Reader
_GOOGLE_API_KEY: str = "google_api_key"


class AssetPairAndHistoricalPrice(NamedTuple):
    from_asset: str
    to_asset: str
    exchange: str
    historical_data: Optional[HistoricalBar] = None


class PairConverterPlugin(AbstractPairConverterPlugin):
    def __init__(
        self,
        historical_price_type: str,
        default_exchange: Optional[str] = None,
        fiat_priority: Optional[str] = None,
        google_api_key: Optional[str] = None,
        exchange_locked: Optional[bool] = None,
    ) -> None:

        super().__init__(historical_price_type=historical_price_type, fiat_priority=fiat_priority)
        self.__logger: logging.Logger = create_logger(f"{self.name()}/{historical_price_type}")

        self.__exchanges: Dict[str, Exchange] = {}
        self.__exchange_markets: Dict[str, Dict[str, List[str]]] = {}
        self.__google_api_key: Optional[str] = google_api_key
        self.__exchange_locked: bool = exchange_locked if exchange_locked is not None else False

        # TO BE IMPLEMENTED - graph and vertex classes to make this more understandable
        # https://github.com/eprbell/dali-rp2/pull/53#discussion_r924056308
        self.__default_exchange: str = _DEFAULT_EXCHANGE if default_exchange is None else default_exchange
        self.__exchange_csv_reader: Dict[str, Any] = {}
        self.__exchange_graphs: Dict[str, Dict[str, Dict[str, None]]] = {}
        self.__exchange_last_request: Dict[str, float] = {}
        if exchange_locked:
            self.__logger.debug("Routing locked to single exchange %s.", self.__default_exchange)
        else:
            self.__logger.debug("Default exchange assigned as %s. _DEFAULT_EXCHANGE is %s", self.__default_exchange, _DEFAULT_EXCHANGE)

    def name(self) -> str:
        return "CCXT-converter"

    def cache_key(self) -> str:
        return self.name()

    @property
    def exchanges(self) -> Dict[str, Exchange]:
        return self.__exchanges

    @property
    def exchange_markets(self) -> Dict[str, Dict[str, List[str]]]:
        return self.__exchange_markets

    @property
    def exchange_graphs(self) -> Dict[str, Dict[str, Dict[str, None]]]:
        return self.__exchange_graphs

    def _bfs_cyclic(self, graph: Dict[str, Dict[str, None]], start: str, end: str) -> Optional[List[str]]:

        # maintain a queue of paths
        # TO BE IMPLEMENTED - using on vertex queue and one dict?
        # https://github.com/eprbell/dali-rp2/pull/53#discussion_r924058754
        queue: List[List[str]] = []
        visited: Dict[str, None] = {}

        # push the first path into the queue
        queue.append([start])

        while queue:
            # get the first path from the queue
            path: List[str] = queue.pop(0)

            # get the last node from the path
            node: str = path[-1]

            # path found
            if node == end:
                return path

            # enumerate all adjacent nodes, construct a new path and push it into the queue
            for adjacent in graph.get(node, {}):

                # prevents an infinite loop.
                if adjacent not in visited:
                    new_path: List[str] = list(path)
                    new_path.append(adjacent)
                    queue.append(new_path)
                    visited[adjacent] = None

        # No path found
        return None

    def _prioritize_quote_assets(self, current_graph: Dict[str, List[str]]) -> None:
        for base_asset in current_graph.keys():
            for priority_quote in reversed(_QUOTE_PRIORITY):
                if priority_quote in current_graph[base_asset]:
                    current_graph[base_asset].pop(current_graph[base_asset].index(priority_quote))
                    remainder: List[str] = current_graph[base_asset]
                    current_graph[base_asset] = [priority_quote]
                    current_graph[base_asset].extend(remainder)

    def get_historic_bar_from_native_source(self, timestamp: datetime, from_asset: str, to_asset: str, exchange: str) -> Optional[HistoricalBar]:
        self.__logger.debug("Converting %s to %s", from_asset, to_asset)

        # If both assets are fiat, skip further processing
        if self._is_fiat_pair(from_asset, to_asset):
            return self._get_fiat_exchange_rate(timestamp, from_asset, to_asset)

        if exchange == Keyword.UNKNOWN.value or exchange not in _EXCHANGE_DICT or self.__exchange_locked:
            if self.__exchange_locked:
                self.__logger.debug("Price routing locked to %s type for %s.", self.__default_exchange, exchange)
            else:
                self.__logger.debug("Using default exchange %s type for %s.", self.__default_exchange, exchange)
            exchange = self.__default_exchange

        # The exchange could have been added as an alt; if so markets wouldn't have been built
        if exchange not in self.__exchanges or exchange not in self.__exchange_markets:
            if self.__exchange_locked:
                self._add_exchange_to_memcache(self.__default_exchange)
            elif exchange in _EXCHANGE_DICT:
                self._add_exchange_to_memcache(exchange)
            else:
                self.__logger.error("WARNING: Unrecognized Exchange: %s. Please open an issue at %s", exchange, self.issues_url)
                return None

        current_markets = self.__exchange_markets[exchange]
        current_graph = self.__exchange_graphs[exchange]

        market_symbol = from_asset + to_asset
        result: Optional[HistoricalBar] = None

        # TO BE IMPLEMENTED - bypass routing if conversion can be done with one market on the exchange
        if market_symbol in current_markets and (exchange in current_markets[market_symbol]):
            self.__logger.debug("Found market - %s on single exchange, skipping routing.", market_symbol)
            result = self.find_historical_bar(from_asset, to_asset, timestamp, exchange)
            return result
        # else:
        # Graph building goes here.

        pricing_path: Optional[List[str]] = self._bfs_cyclic(current_graph, from_asset, to_asset)
        if pricing_path is None:
            self.__logger.debug("No path found for %s to %s. Please open an issue at %s.", from_asset, to_asset, self.issues_url)
            return None

        self.__logger.debug("Found path - %s", pricing_path)

        conversion_route: List[AssetPairAndHistoricalPrice] = []
        last_node: Optional[str] = None
        hop_bar: Optional[HistoricalBar] = None

        # Build conversion stack, we will iterate over this to find the price for each conversion
        # Then multiply them together to get our final price.
        for node in pricing_path:

            if last_node:
                conversion_route.append(
                    AssetPairAndHistoricalPrice(
                        from_asset=last_node,
                        to_asset=node,
                        exchange=current_markets[(last_node + node)][0],
                        historical_data=None,
                    )
                )

            last_node = node

        for i, hop_data in enumerate(conversion_route):
            if self._is_fiat_pair(hop_data.from_asset, hop_data.to_asset):
                hop_bar = self._get_fiat_exchange_rate(timestamp, hop_data.from_asset, hop_data.to_asset)
            else:
                hop_bar = self.find_historical_bar(hop_data.from_asset, hop_data.to_asset, timestamp, hop_data.exchange)

            if hop_bar is not None:
                # Replacing an immutable attribute
                conversion_route[i] = conversion_route[i]._replace(historical_data=hop_bar)
            else:
                self.__logger.debug(
                    """No pricing data found for hop. This could be caused by airdropped
                    coins that do not have a market yet. Market - %s%s, Timestamp - %s, Exchange - %s""",
                    hop_data.from_asset,
                    hop_data.to_asset,
                    timestamp,
                    hop_data.exchange,
                )

            if result is not None:
                # TO BE IMPLEMENTED - override Historical Bar * to multiply two bars?
                result = HistoricalBar(
                    duration=max(result.duration, hop_bar.duration),  # type: ignore
                    timestamp=timestamp,
                    open=(result.open * hop_bar.open),  # type: ignore
                    high=(result.high * hop_bar.high),  # type: ignore
                    low=(result.low * hop_bar.low),  # type: ignore
                    close=(result.close * hop_bar.close),  # type: ignore
                    volume=(result.volume + hop_bar.volume),  # type: ignore
                )
            else:
                result = hop_bar

        return result

    def find_historical_bar(self, from_asset: str, to_asset: str, timestamp: datetime, exchange: str) -> Optional[HistoricalBar]:
        result: Optional[HistoricalBar] = None
        retry_count: int = 0
        current_exchange: Any = self.__exchanges[exchange]
        ms_timestamp: int = int(timestamp.timestamp() * _MS_IN_SECOND)
        key: AssetPairAndTimestamp = AssetPairAndTimestamp(timestamp, from_asset, to_asset, exchange)
        historical_bar: Optional[HistoricalBar] = self._get_bar_from_cache(key)
        csv_pricing: Any = _CSV_PRICING_DICT.get(exchange)
        csv_reader: Any = None

        if historical_bar is not None:
            self.__logger.debug("Retrieved cache for %s/%s->%s for %s", timestamp, from_asset, to_asset, exchange)
            return historical_bar

        if self.__exchange_csv_reader.get(exchange):
            csv_reader = self.__exchange_csv_reader[exchange]
        elif csv_pricing is not None:
            csv_signature: Signature = signature(csv_pricing)

            # a Google API key is necessary to interact with Google Drive since Google restricts API calls to avoid spam, etc...
            if _GOOGLE_API_KEY in csv_signature.parameters:
                if self.__google_api_key is not None:
                    csv_reader = csv_pricing(self.__google_api_key)
                else:
                    self.__logger.info(
                        "Google API Key is not set. Setting the Google API key in the CCXT pair converter plugin could speed up pricing resolution"
                    )
            else:
                csv_reader = csv_pricing()

        if csv_reader:
            csv_bar: Optional[HistoricalBar] = csv_reader.find_historical_bar(from_asset, to_asset, timestamp)

            if csv_bar:
                self._add_bar_to_cache(key=AssetPairAndTimestamp(timestamp, from_asset, to_asset, exchange), historical_bar=csv_bar)

            historical_bar = self._get_bar_from_cache(key)
            self.__exchange_csv_reader[exchange] = csv_reader
            if historical_bar is not None:
                self.__logger.debug("Retrieved bar cache - %s for %s/%s->%s for %s", historical_bar, key.timestamp, key.from_asset, key.to_asset, key.exchange)
                return historical_bar

        while retry_count < len(_TIME_GRANULARITY):

            timeframe: str = _TIME_GRANULARITY[retry_count]
            request_count: int = 0
            historical_data: List[List[Union[int, float]]] = []

            # Most exceptions are caused by request limits of the underlying APIs
            while request_count < 9:
                try:
                    # Excessive calls to the API within a certain window might get an IP temporarily banned
                    if _REQUEST_DELAYDICT.get(exchange, 0) > 0:
                        current_time = time()
                        second_delay = max(0, _REQUEST_DELAYDICT[exchange] - (current_time - self.__exchange_last_request.get(exchange, 0)))
                        self.__logger.debug("Delaying for %s seconds", second_delay)
                        sleep(second_delay)
                        self.__exchange_last_request[exchange] = time()

                    # this is where we pull the historical prices from the underlying exchange
                    historical_data = current_exchange.fetchOHLCV(f"{from_asset}/{to_asset}", timeframe, ms_timestamp, 1)
                    break
                except (DDoSProtection, ExchangeError) as exc:
                    self.__logger.debug(
                        "Exception from server, most likely too many requests. Making another attempt after 0.1 second delay. Exception - %s", exc
                    )
                    # logger INFO for retry?
                    sleep(0.1)
                    request_count += 3
                except (ExchangeNotAvailable, NetworkError, RequestTimeout) as exc_na:
                    request_count += 1
                    if request_count > 9:
                        if exchange == _BINANCE:
                            self.__logger.info(
                                """
                                Binance server unavailable. Try a non-Binance locked exchange pair converter.
                                Saving to cache and exiting.
                                """
                            )
                        else:
                            self.__logger.info("Maximum number of retries reached. Saving to cache and exiting.")
                        self.save_historical_price_cache()
                        raise RP2RuntimeError("Server error") from exc_na

                    self.__logger.debug("Server not available. Making attempt #%s of 10 after a ten second delay. Exception - %s", request_count, exc_na)
                    sleep(10)

            # If there is no candle the list will be empty
            if historical_data:
                result = HistoricalBar(
                    duration=timedelta(seconds=_TIME_GRANULARITY_IN_SECONDS[retry_count]),
                    timestamp=timestamp,
                    open=RP2Decimal(str(historical_data[0][1])),
                    high=RP2Decimal(str(historical_data[0][2])),
                    low=RP2Decimal(str(historical_data[0][3])),
                    close=RP2Decimal(str(historical_data[0][4])),
                    volume=RP2Decimal(str(historical_data[0][5])),
                )
                break

            retry_count += 1

        # Save the individual pair to cache
        if result is not None:
            self._add_bar_to_cache(key, result)

        return result

    def _add_alternative_markets(self, current_graph: Dict[str, Dict[str, None]], current_markets: Dict[str, List[str]]) -> None:
        for base_asset, quote_asset in _ALT_MARKET_BY_BASE_DICT.items():
            alt_market = base_asset + quote_asset
            alt_exchange_name = _ALT_MARKET_EXCHANGES_DICT[alt_market]

            # TO BE IMPLEMENTED - Add alt market to the end of list if another exchange exists already
            current_markets[alt_market] = [alt_exchange_name]

            # Cache the exchange so that we can pull prices from it later
            if alt_exchange_name not in self.__exchanges:
                self.__logger.debug("Added Alternative Exchange: %s", alt_exchange_name)
                alt_exchange: Exchange = _EXCHANGE_DICT[alt_exchange_name]()
                self.__exchanges[alt_exchange_name] = alt_exchange

            if current_graph.get(base_asset) and (quote_asset not in current_graph[base_asset]):
                current_graph[base_asset][quote_asset] = None
            else:
                current_graph[base_asset] = {quote_asset: None}

    def _add_exchange_to_memcache(self, exchange: str) -> None:

        if exchange not in self.__exchanges:
            # initializes the cctx exchange instance which is used to get the historical data
            # https://docs.ccxt.com/en/latest/manual.html#notes-on-rate-limiter
            current_exchange: Exchange = _EXCHANGE_DICT[exchange]({"enableRateLimit": True})
        else:
            current_exchange = self.__exchanges[exchange]

        # key: market, value: exchanges where the market is available in order of priority
        current_markets: Dict[str, List[str]] = {}
        current_graph: Dict[str, Dict[str, None]] = {}

        for market in filter(lambda x: x[_TYPE] == "spot" and x[_QUOTE] in _QUOTE_PRIORITY, current_exchange.fetch_markets()):
            self.__logger.debug("Market: %s", market)

            current_markets[f"{market[_BASE]}{market[_QUOTE]}"] = [exchange]

            # TO BE IMPLEMENTED - lazy build graph only if needed

            # Add the quote asset to the graph if it isn't there already.
            current_graph.setdefault(market[_BASE], {})[market[_QUOTE]] = None

        self._prioritize_quote_assets(current_markets)

        # Add alternative markets if they don't exist
        if not self.__exchange_locked:
            self._add_alternative_markets(current_graph, current_markets)

        self._add_fiat_edges_to_graph(current_graph, current_markets)
        self.__logger.debug("Added graph for %s : %s", current_exchange, current_graph)
        self.__exchanges[exchange] = current_exchange
        self.__exchange_markets[exchange] = current_markets
        self.__exchange_graphs[exchange] = current_graph



# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# CCXT documentation:
# https://docs.ccxt.com/en/latest/index.html

# pylint: disable=fixme

import json
import logging
from datetime import datetime, timezone
from multiprocessing.pool import ThreadPool
from time import sleep
from typing import Any, Callable, Dict, Iterable, List, NamedTuple, Optional, Union

from ccxt import (
    DDoSProtection,
    Exchange,
    ExchangeError,
    ExchangeNotAvailable,
    NetworkError,
    RequestTimeout,
)
from rp2.abstract_country import AbstractCountry
from rp2.logger import create_logger
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2RuntimeError

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.ccxt_pagination import (
    AbstractPaginationDetailSet,
    AbstractPaginationDetailsIterator,
    PaginationDetails,
)
from dali.configuration import Keyword
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.out_transaction import OutTransaction

# Native format keywords
_AMOUNT: str = "amount"
_BUY: str = "buy"
_COST: str = "cost"
_CURRENCY: str = "currency"
_DATE_TIME: str = "datetime"
_DEPOSIT: str = "deposit"
_FEE: str = "fee"
_FETCH_DEPOSITS: str = "fetchDeposits"
_FETCH_MY_TRADES: str = "fetchMyTrades"
_FETCH_WITHDRAWALS: str = "fetchWithdrawals"
_ID: str = "id"
_PRICE: str = "price"
_SELL: str = "sell"
_SIDE: str = "side"
_STATUS: str = "status"
_SYMBOL: str = "symbol"
_TIMESTAMP: str = "timestamp"
_TYPE: str = "type"
_TX_ID: str = "txid"  # CCXT doesn't capitalize I
_WITHDRAWAL: str = "withdrawal"

_MS_IN_SECOND: int = 1000


class Trade(NamedTuple):
    base_asset: str
    quote_asset: str
    base_info: str
    quote_info: str


class ProcessOperationResult(NamedTuple):
    in_transactions: List[InTransaction]
    out_transactions: List[OutTransaction]
    intra_transactions: List[IntraTransaction]


class AbstractCcxtInputPlugin(AbstractInputPlugin):

    __DEFAULT_THREAD_COUNT: int = 1

    def __init__(
        self,
        account_holder: str,
        exchange_start_time: datetime,
        native_fiat: Optional[str],
        thread_count: Optional[int],
    ) -> None:

        super().__init__(account_holder, native_fiat)
        self.__logger: logging.Logger = create_logger(f"{self.exchange_name()}/{self.account_holder}")
        self.__cache_key: str = f"{str(self.exchange_name()).lower()}-{account_holder}"
        self.__client: Exchange = self._initialize_client()
        self.__thread_count = thread_count if thread_count else self.__DEFAULT_THREAD_COUNT
        self.__markets: List[str] = []
        self.__start_time: datetime = exchange_start_time
        self.__start_time_ms: int = int(self.__start_time.timestamp()) * _MS_IN_SECOND

    def plugin_name(self) -> str:
        raise NotImplementedError("Abstract method")

    def cache_key(self) -> Optional[str]:
        return self.__cache_key

    def _initialize_client(self) -> Exchange:
        raise NotImplementedError("Abstract method")

    def exchange_name(self) -> str:
        raise NotImplementedError("Abstract method")

    def _get_process_deposits_pagination_detail_set(self) -> Optional[AbstractPaginationDetailSet]:
        raise NotImplementedError("Abstract method")

    def _get_process_withdrawals_pagination_detail_set(self) -> Optional[AbstractPaginationDetailSet]:
        raise NotImplementedError("Abstract method")

    # Some exchanges require you to loop through all markets for trades
    def _get_process_trades_pagination_detail_set(self) -> Optional[AbstractPaginationDetailSet]:
        raise NotImplementedError("Abstract method")

    def _get_markets(self) -> List[str]:

        if self.__markets:
            return self.__markets

        ccxt_markets: Any = self._client.fetch_markets()
        market_list: List[str] = []
        for market in ccxt_markets:
            self.__logger.debug("Market: %s", json.dumps(market))
            if market[_TYPE] == "spot":
                market_list.append(market[_ID])

        self.__markets = market_list

        return self.__markets

    @staticmethod
    def _rp2_timestamp_from_ms_epoch(epoch_timestamp: str) -> str:
        rp2_time = datetime.fromtimestamp((int(epoch_timestamp) / _MS_IN_SECOND), timezone.utc)

        return rp2_time.strftime("%Y-%m-%d %H:%M:%S%z")

    # Parses the symbol (eg. 'BTC/USD') into base and quote assets, and formats notes for the transactions
    @staticmethod
    def _to_trade(market_pair: str, base_amount: str, quote_amount: str) -> Trade:
        assets = market_pair.split("/")
        return Trade(
            base_asset=assets[0],
            quote_asset=assets[1],
            base_info=f"{base_amount} {assets[0]}",
            quote_info=f"{quote_amount} {assets[1]}",
        )

    @property
    def _client(self) -> Exchange:
        return self.__client

    @property
    def _logger(self) -> logging.Logger:
        return self.__logger

    @property
    def _start_time_ms(self) -> int:
        return self.__start_time_ms

    @property
    def _thread_count(self) -> int:
        return self.__thread_count

    def load(self, country: AbstractCountry) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []
        in_transactions: List[InTransaction] = []
        out_transactions: List[OutTransaction] = []
        intra_transactions: List[IntraTransaction] = []

        if self._client.has[_FETCH_DEPOSITS]:
            self._process_deposits(intra_transactions)
        self._process_gains(in_transactions, out_transactions)
        if self._client.has[_FETCH_MY_TRADES]:
            self._process_trades(in_transactions, out_transactions)
        if self._client.has[_FETCH_WITHDRAWALS]:
            self._process_withdrawals(intra_transactions)
        self._process_implicit_api(in_transactions, out_transactions, intra_transactions)

        result.extend(in_transactions)
        result.extend(out_transactions)
        result.extend(intra_transactions)

        return result

    ### Multiple Transaction Processing or Pagination Methods

    def _process_deposits(
        self,
        intra_transactions: List[IntraTransaction],
    ) -> None:
        processing_result_list: List[Optional[ProcessOperationResult]] = []
        pagination_detail_set: Optional[AbstractPaginationDetailSet] = self._get_process_deposits_pagination_detail_set()
        # Strip optionality
        if not pagination_detail_set:
            raise RP2RuntimeError("No Pagination Details for Deposits")

        has_pagination_detail_set: AbstractPaginationDetailSet = pagination_detail_set

        pagination_detail_iterator: AbstractPaginationDetailsIterator = iter(has_pagination_detail_set)

        try:
            while True:
                pagination_details: PaginationDetails = next(pagination_detail_iterator)
                deposits = self.__safe_api_call(
                    self._client.fetch_deposits,
                    {
                        "code": pagination_details.symbol,
                        "since": pagination_details.since,
                        "limit": pagination_details.limit,
                        "params": pagination_details.params,
                    },
                )
                # CCXT returns a standardized response from fetch_deposits. 'info' is the exchange-specific information
                # in this case from Binance.com

                # {
                #   'info': {
                #       'amount': '0.00999800',
                #       'coin': 'PAXG',
                #       'network': 'ETH',
                #       'status': '1',
                #       'address': '0x788cabe9236ce061e5a892e1a59395a81fc8d62c',
                #       'addressTag': '',
                #       'txId': '0xaad4654a3234aa6118af9b4b335f5ae81c360b2394721c019b5d1e75328b09f3',
                #       'insertTime': '1599621997000',
                #       'transferType': '0',
                #       'confirmTimes': '12/12',
                #       'unlockConfirm': '12/12',
                #       'walletType': '0'
                #   },
                #   'id': None,
                #   'txid': '0xaad4654a3234aa6118af9b4b335f5ae81c360b2394721c019b5d1e75328b09f3',
                #   'timestamp': 1599621997000,
                #   'datetime': '2020-09-09T03:26:37.000Z',
                #   'network': 'ETH',
                #   'address': '0x788cabe9236ce061e5a892e1a59395a81fc8d62c',
                #   'addressTo': '0x788cabe9236ce061e5a892e1a59395a81fc8d62c',
                #   'addressFrom': None,
                #   'tag': None,
                #   'tagTo': None,
                #   'tagFrom': None,
                #   'type': 'deposit',
                #   'amount': 0.00999800,
                #   'currency': 'PAXG',
                #   'status': 'ok',
                #   'updated': None,
                #   'internal': False,
                #   'fee': None
                # }

                pagination_detail_iterator.update_fetched_elements(deposits)

                with ThreadPool(self._thread_count) as pool:
                    processing_result_list = pool.map(self._process_transfer, deposits)

                for processing_result in processing_result_list:
                    if processing_result is None:
                        continue
                    if processing_result.intra_transactions:
                        intra_transactions.extend(processing_result.intra_transactions)

        except StopIteration:
            # End of pagination details
            pass

    def _process_gains(
        self,
        in_transactions: List[InTransaction],
        out_transactions: List[OutTransaction],
    ) -> None:
        raise NotImplementedError("Abstract method")

    def _process_implicit_api(
        self,
        in_transactions: List[InTransaction],
        out_transactions: List[OutTransaction],
        intra_transactions: List[IntraTransaction],
    ) -> None:
        raise NotImplementedError("Abstract method")

    def _process_trades(
        self,
        in_transactions: List[InTransaction],
        out_transactions: List[OutTransaction],
    ) -> None:

        processing_result_list: List[Optional[ProcessOperationResult]] = []
        pagination_detail_set: Optional[AbstractPaginationDetailSet] = self._get_process_trades_pagination_detail_set()
        # Strip optionality
        if not pagination_detail_set:
            raise RP2RuntimeError("No pagination details for trades.")

        has_pagination_detail_set: AbstractPaginationDetailSet = pagination_detail_set

        pagination_detail_iterator: AbstractPaginationDetailsIterator = iter(has_pagination_detail_set)
        try:
            while True:
                pagination_details: PaginationDetails = next(pagination_detail_iterator)

                trades: Iterable[Dict[str, Union[str, float]]] = self.__safe_api_call(
                    self._client.fetch_my_trades,
                    {
                        "symbol": pagination_details.symbol,
                        "since": pagination_details.since,
                        "limit": pagination_details.limit,
                        "params": pagination_details.params,
                    },
                )
                #   {
                #       'info':         { ... },                    // the original decoded JSON as is
                #       'id':           '12345-67890:09876/54321',  // string trade id
                #       'timestamp':    1502962946216,              // Unix timestamp in milliseconds
                #       'datetime':     '2017-08-17 12:42:48.000',  // ISO8601 datetime with milliseconds
                #       'symbol':       'ETH/BTC',                  // symbol
                #       'order':        '12345-67890:09876/54321',  // string order id or undefined/None/null
                #       'type':         'limit',                    // order type, 'market', 'limit' or undefined/None/null
                #       'side':         'buy',                      // direction of the trade, 'buy' or 'sell'
                #       'takerOrMaker': 'taker',                    // string, 'taker' or 'maker'
                #       'price':        0.06917684,                 // float price in quote currency
                #       'amount':       1.5,                        // amount of base currency
                #       'cost':         0.10376526,                 // total cost, `price * amount`,
                #       'fee':          {                           // provided by exchange or calculated by ccxt
                #           'cost':  0.0015,                        // float
                #           'currency': 'ETH',                      // usually base currency for buys, quote currency for sells
                #           'rate': 0.002,                          // the fee rate (if available)
                #       },
                #   }

                # * The ``fee`` currency may be different from both traded currencies (for example, an ETH/BTC order with fees in USD).
                # * The ``cost`` of the trade means ``amount * price``. It is the total *quote* volume of the trade (whereas `amount` is the *base* volume).
                # * The cost field itself is there mostly for convenience and can be deduced from other fields.
                # * The ``cost`` of the trade is a *"gross"* value. That is the value pre-fee, and the fee has to be applied afterwards.

                pagination_detail_iterator.update_fetched_elements(trades)

                with ThreadPool(self._thread_count) as pool:
                    processing_result_list = pool.map(self._process_buy_and_sell, trades)

                for processing_result in processing_result_list:
                    if processing_result is None:
                        continue
                    if processing_result.in_transactions:
                        in_transactions.extend(processing_result.in_transactions)
                    if processing_result.out_transactions:
                        out_transactions.extend(processing_result.out_transactions)

        except StopIteration:
            # End of pagination details
            pass

    def _process_withdrawals(
        self,
        intra_transactions: List[IntraTransaction],
    ) -> None:

        processing_result_list: List[Optional[ProcessOperationResult]] = []
        pagination_detail_set: Optional[AbstractPaginationDetailSet] = self._get_process_withdrawals_pagination_detail_set()
        # Strip optionality
        if not pagination_detail_set:
            raise RP2RuntimeError("No pagination details for withdrawals.")

        has_pagination_detail_set: AbstractPaginationDetailSet = pagination_detail_set

        pagination_detail_iterator: AbstractPaginationDetailsIterator = iter(has_pagination_detail_set)

        try:
            while True:
                pagination_details: PaginationDetails = next(pagination_detail_iterator)
                withdrawals: Iterable[Dict[str, Union[str, float]]] = self.__safe_api_call(
                    self._client.fetch_withdrawals,
                    {
                        "code": pagination_details.symbol,
                        "since": pagination_details.since,
                        "limit": pagination_details.limit,
                        "params": pagination_details.params,
                    },
                )
                # {
                #   'info': {
                #       'amount': '0.00999800',
                #       'coin': 'PAXG',
                #       'network': 'ETH',
                #       'status': '1',
                #       'address': '0x788cabe9236ce061e5a892e1a59395a81fc8d62c',
                #       'addressTag': '',
                #       'txId': '0xaad4654a3234aa6118af9b4b335f5ae81c360b2394721c019b5d1e75328b09f3',
                #       'insertTime': '1599621997000',
                #       'transferType': '0',
                #       'confirmTimes': '12/12',
                #       'unlockConfirm': '12/12',
                #       'walletType': '0'
                #   },
                #   'id': None,
                #   'txid': '0xaad4654a3234aa6118af9b4b335f5ae81c360b2394721c019b5d1e75328b09f3',
                #   'timestamp': 1599621997000,
                #   'datetime': '2020-09-09T03:26:37.000Z',
                #   'network': 'ETH',
                #   'address': '0x788cabe9236ce061e5a892e1a59395a81fc8d62c',
                #   'addressTo': '0x788cabe9236ce061e5a892e1a59395a81fc8d62c',
                #   'addressFrom': None,
                #   'tag': None,
                #   'tagTo': None,
                #   'tagFrom': None,
                #   'type': 'withdrawal',
                #   'amount': 0.00999800,
                #   'currency': 'PAXG',
                #   'status': 'ok',
                #   'updated': None,
                #   'internal': False,
                #   'fee': None
                # }
                pagination_detail_iterator.update_fetched_elements(withdrawals)

                with ThreadPool(self._thread_count) as pool:
                    processing_result_list = pool.map(self._process_transfer, withdrawals)

                for processing_result in processing_result_list:
                    if processing_result is None:
                        continue
                    if processing_result.intra_transactions:
                        intra_transactions.extend(processing_result.intra_transactions)

        except StopIteration:
            # End of pagination details
            pass

    def __safe_api_call(
        self,
        function: Callable[..., Iterable[Dict[str, Union[str, float]]]],
        params: Dict[str, Any],
    ) -> Iterable[Dict[str, Union[str, float]]]:

        results: Iterable[Dict[str, Union[str, float]]] = {}
        request_count: int = 0

        # Most exceptions are caused by request limits of the underlying APIs
        while request_count < 9:
            try:
                if "code" in params:
                    results = function(**params)
                else:
                    results = function(**params)
                break
            except (DDoSProtection, ExchangeError) as exc:
                self.__logger.debug("Exception from server, most likely too many requests. Making another attempt after 0.1 second delay. Exception - %s", exc)
                sleep(0.1)
                request_count += 3
            except (ExchangeNotAvailable, NetworkError, RequestTimeout) as exc_na:
                request_count += 1
                if request_count > 9:
                    self.__logger.info("Maximum number of retries reached.")
                    raise RP2RuntimeError("Server error") from exc_na

                self.__logger.debug("Server not available. Making attempt #%s of 10 after a ten second delay. Exception - %s", request_count, exc_na)
                sleep(10)

        return results

    ### Single Transaction Processing

    def _process_buy_and_sell(self, transaction: Any, notes: Optional[str] = None) -> ProcessOperationResult:
        results: ProcessOperationResult = self._process_buy(transaction, notes)
        results.out_transactions.extend(self._process_sell(transaction, notes).out_transactions)

        return results

    def _process_buy(self, transaction: Any, notes: Optional[str] = None) -> ProcessOperationResult:
        self.__logger.debug("Buy: %s", json.dumps(transaction))
        in_transaction_list: List[InTransaction] = []
        out_transaction_list: List[OutTransaction] = []
        crypto_in: RP2Decimal
        crypto_fee: RP2Decimal
        fee_asset: str = transaction[_FEE][_CURRENCY]

        trade: Trade = self._to_trade(transaction[_SYMBOL], str(transaction[_AMOUNT]), str(transaction[_COST]))
        if transaction[_SIDE] == _BUY:
            out_asset = trade.quote_asset
            in_asset = trade.base_asset
            crypto_in = RP2Decimal(str(transaction[_AMOUNT]))
            conversion_info = f"{trade.quote_info} -> {trade.base_info}"
        elif transaction[_SIDE] == _SELL:
            out_asset = trade.base_asset
            in_asset = trade.quote_asset
            crypto_in = RP2Decimal(str(transaction[_COST]))
            conversion_info = f"{trade.base_info} -> {trade.quote_info}"
        else:
            raise RP2RuntimeError(f"Internal error: unrecognized transaction side: {transaction[_SIDE]}")

        if fee_asset == in_asset:
            crypto_fee = RP2Decimal(str(transaction[_FEE][_COST]))
        else:
            crypto_fee = ZERO
            transaction_fee = RP2Decimal(str(transaction[_FEE][_COST]))

            # Users can use other crypto assets to pay for trades
            if fee_asset != out_asset and RP2Decimal(transaction_fee) > ZERO:
                out_transaction_list.append(
                    OutTransaction(
                        plugin=self.plugin_name(),
                        unique_id=transaction[_ID],
                        raw_data=json.dumps(transaction),
                        timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_TIMESTAMP]),
                        asset=transaction[_FEE][_CURRENCY],
                        exchange=self.exchange_name(),
                        holder=self.account_holder,
                        transaction_type=Keyword.FEE.value,
                        spot_price=Keyword.UNKNOWN.value,
                        crypto_out_no_fee="0",
                        crypto_fee=str(transaction_fee),
                        crypto_out_with_fee=str(transaction_fee),
                        fiat_out_no_fee=str(transaction_fee) if self.is_native_fiat(fee_asset) else None,
                        fiat_fee=str(transaction_fee) if self.is_native_fiat(fee_asset) else None,
                        notes=(f"{notes + '; ' if notes else ''} Fee for conversion from " f"{conversion_info}"),
                    )
                )

        # Is this a plain buy or a conversion?
        if self.is_native_fiat(trade.quote_asset):
            fiat_in_with_fee = RP2Decimal(str(transaction[_COST]))
            fiat_fee = RP2Decimal(crypto_fee)
            spot_price = RP2Decimal(str(transaction[_PRICE]))
            if transaction[_SIDE] == _BUY:
                transaction_notes = f"Fiat buy of {trade.base_asset} with {trade.quote_asset}"
                fiat_in_no_fee = fiat_in_with_fee - (fiat_fee * spot_price)
            elif transaction[_SIDE] == _SELL:
                transaction_notes = f"Fiat sell of {trade.base_asset} into {trade.quote_asset}"
                fiat_in_no_fee = fiat_in_with_fee - fiat_fee

            in_transaction_list.append(
                InTransaction(
                    plugin=self.plugin_name(),
                    unique_id=transaction[_ID],
                    raw_data=json.dumps(transaction),
                    timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_TIMESTAMP]),
                    asset=in_asset,
                    exchange=self.exchange_name(),
                    holder=self.account_holder,
                    transaction_type=Keyword.BUY.value,
                    spot_price=str(spot_price),
                    crypto_in=str(crypto_in),
                    crypto_fee=None if self.is_native_fiat(in_asset) else str(crypto_fee),
                    fiat_in_no_fee=str(fiat_in_no_fee),
                    fiat_in_with_fee=str(fiat_in_with_fee),
                    fiat_fee=str(fiat_fee) if self.is_native_fiat(in_asset) else None,
                    fiat_ticker=trade.quote_asset,
                    notes=(f"{notes + '; ' if notes else ''} {transaction_notes}"),
                )
            )

        else:
            transaction_notes = f"Buy side of conversion from " f"{conversion_info}" f"({out_asset} out-transaction unique id: {transaction[_ID]}"

            in_transaction_list.append(
                InTransaction(
                    plugin=self.plugin_name(),
                    unique_id=transaction[_ID],
                    raw_data=json.dumps(transaction),
                    timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_TIMESTAMP]),
                    asset=in_asset,
                    exchange=self.exchange_name(),
                    holder=self.account_holder,
                    transaction_type=Keyword.BUY.value,
                    spot_price=Keyword.UNKNOWN.value,
                    crypto_in=str(crypto_in),
                    crypto_fee=str(crypto_fee),
                    fiat_in_no_fee=None,
                    fiat_in_with_fee=None,
                    fiat_fee=None,
                    notes=(f"{notes + '; ' if notes else ''} {transaction_notes}"),
                )
            )

        return ProcessOperationResult(in_transactions=in_transaction_list, out_transactions=out_transaction_list, intra_transactions=[])

    def _process_sell(self, transaction: Any, notes: Optional[str] = None) -> ProcessOperationResult:
        self.__logger.debug("Sell: %s", json.dumps(transaction))
        out_transaction_list: List[OutTransaction] = []
        trade: Trade = self._to_trade(transaction[_SYMBOL], str(transaction[_AMOUNT]), str(transaction[_COST]))

        # For some reason CCXT outputs amounts in float
        if transaction[_SIDE] == _BUY:
            out_asset = trade.quote_asset
            in_asset = trade.base_asset
            crypto_out_no_fee: RP2Decimal = RP2Decimal(str(transaction[_COST]))
            conversion_info = f"{trade.quote_info} -> {trade.base_info}"
        elif transaction[_SIDE] == _SELL:
            out_asset = trade.base_asset
            in_asset = trade.quote_asset
            crypto_out_no_fee = RP2Decimal(str(transaction[_AMOUNT]))
            conversion_info = f"{trade.base_info} -> {trade.quote_info}"
        else:
            raise RP2RuntimeError(f"Internal error: unrecognized transaction side: {transaction[_SIDE]}")

        if transaction[_FEE][_CURRENCY] == out_asset:
            crypto_fee: RP2Decimal = RP2Decimal(str(transaction[_FEE][_COST]))
        else:
            crypto_fee = ZERO
        crypto_out_with_fee: RP2Decimal = crypto_out_no_fee + crypto_fee

        # Is this a plain buy or a conversion?
        if self.is_native_fiat(trade.quote_asset):
            fiat_out_no_fee: RP2Decimal = RP2Decimal(str(transaction[_COST]))
            fiat_fee: RP2Decimal = crypto_fee
            spot_price: RP2Decimal = RP2Decimal(str(transaction[_PRICE]))

            out_transaction_list.append(
                OutTransaction(
                    plugin=self.plugin_name(),
                    unique_id=transaction[_ID],
                    raw_data=json.dumps(transaction),
                    timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_TIMESTAMP]),
                    asset=out_asset,
                    exchange=self.exchange_name(),
                    holder=self.account_holder,
                    transaction_type=Keyword.SELL.value,
                    spot_price=str(spot_price),
                    crypto_out_no_fee=str(crypto_out_no_fee),
                    crypto_fee=str(crypto_fee),
                    crypto_out_with_fee=str(crypto_out_with_fee),
                    fiat_out_no_fee=str(fiat_out_no_fee),
                    fiat_fee=str(fiat_fee),
                    fiat_ticker=trade.quote_asset,
                    notes=(f"{notes + ';' if notes else ''} Fiat sell of {trade.base_asset} with {trade.quote_asset}."),
                )
            )

        else:
            # CCXT does not report the value of the transaction in fiat
            out_transaction_list.append(
                OutTransaction(
                    plugin=self.plugin_name(),
                    unique_id=transaction[_ID],
                    raw_data=json.dumps(transaction),
                    timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_TIMESTAMP]),
                    asset=out_asset,
                    exchange=self.exchange_name(),
                    holder=self.account_holder,
                    transaction_type=Keyword.SELL.value,
                    spot_price=Keyword.UNKNOWN.value,
                    crypto_out_no_fee=str(crypto_out_no_fee),
                    crypto_fee=str(crypto_fee),
                    crypto_out_with_fee=str(crypto_out_with_fee),
                    fiat_out_no_fee=None,
                    fiat_fee=None,
                    notes=(
                        f"{notes + '; ' if notes else ''} Sell side of conversion from "
                        f"{conversion_info}"
                        f"({in_asset} in-transaction unique id: {transaction[_ID]}"
                    ),
                )
            )

        return ProcessOperationResult(out_transactions=out_transaction_list, in_transactions=[], intra_transactions=[])

    def _process_transfer(self, transaction: Any) -> ProcessOperationResult:
        self.__logger.debug("Transfer: %s", json.dumps(transaction))
        if transaction[_STATUS] == "failed":
            self.__logger.info("Skipping failed transfer %s", json.dumps(transaction))
        else:
            intra_transaction_list: List[IntraTransaction] = []

            # This is a CCXT list must convert to string from float
            amount: RP2Decimal = RP2Decimal(str(transaction[_AMOUNT]))

            if transaction[_TYPE] == _DEPOSIT:
                intra_transaction_list.append(
                    IntraTransaction(
                        plugin=self.plugin_name(),
                        unique_id=transaction[_TX_ID],
                        raw_data=json.dumps(transaction),
                        timestamp=transaction[_DATE_TIME],
                        asset=transaction[_CURRENCY],
                        from_exchange=Keyword.UNKNOWN.value,
                        from_holder=Keyword.UNKNOWN.value,
                        to_exchange=self.exchange_name(),
                        to_holder=self.account_holder,
                        spot_price=Keyword.UNKNOWN.value,
                        crypto_sent=Keyword.UNKNOWN.value,
                        crypto_received=str(amount),
                    )
                )
            elif transaction[_TYPE] == _WITHDRAWAL:
                intra_transaction_list.append(
                    IntraTransaction(
                        plugin=self.plugin_name(),
                        unique_id=transaction[_TX_ID],
                        raw_data=json.dumps(transaction),
                        timestamp=transaction[_DATE_TIME],
                        asset=transaction[_CURRENCY],
                        from_exchange=self.exchange_name(),
                        from_holder=self.account_holder,
                        to_exchange=Keyword.UNKNOWN.value,
                        to_holder=Keyword.UNKNOWN.value,
                        spot_price=Keyword.UNKNOWN.value,
                        crypto_sent=str(amount),
                        crypto_received=Keyword.UNKNOWN.value,
                    )
                )
            else:
                self.__logger.error("Unrecognized Crypto transfer: %s", json.dumps(transaction))

        return ProcessOperationResult(out_transactions=[], in_transactions=[], intra_transactions=intra_transaction_list)



# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rp2.plugin.country.us import US
from rp2.rp2_decimal import ZERO, RP2Decimal

from dali.configuration import Keyword
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.out_transaction import OutTransaction
from dali.plugin.input.csv.pionex import InputPlugin


class TestPionex:
    def test_trades(self) -> None:
        plugin = InputPlugin(
            account_holder="tester",
            trades_csv_file="input/test_pionex_trades.csv",
            transfers_csv_file=None,
            native_fiat="USD",
        )

        result = plugin.load(US())

        # 1 ETH in transaction for 0.1 ETH +
        # 1 BUSD out transaction for 200 BUSD +
        # 1 BUSD in transaction for 0.05 ETH +
        # 1 ETH out transaction for 110 BUSD = 4
        assert len(result) == 4

        eth_transaction_in: InTransaction = result[0]  # type: ignore
        busd_transaction_out: OutTransaction = result[1]  # type: ignore
        busd_transaction_in: InTransaction = result[2]  # type: ignore
        eth_transaction_out: OutTransaction = result[3]  # type: ignore

        # Buy ETH autoinvest
        assert eth_transaction_in.asset == "ETH"
        assert eth_transaction_in.timestamp == "2022-02-27 12:09:00+0000"
        assert eth_transaction_in.transaction_type == Keyword.BUY.value.capitalize()
        assert eth_transaction_in.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(eth_transaction_in.crypto_in) == RP2Decimal("0.1")
        assert RP2Decimal(str(eth_transaction_in.crypto_fee)) == RP2Decimal("0.00995")
        assert eth_transaction_in.fiat_in_no_fee is None
        assert eth_transaction_in.fiat_in_with_fee is None
        assert eth_transaction_in.fiat_fee is None

        assert busd_transaction_out.asset == "BUSD"
        assert busd_transaction_out.timestamp == "2022-02-27 12:09:00+0000"
        assert busd_transaction_out.transaction_type == Keyword.SELL.value.capitalize()
        assert busd_transaction_out.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(busd_transaction_out.crypto_out_no_fee) == RP2Decimal("200.00")
        assert RP2Decimal(busd_transaction_out.crypto_fee) == ZERO
        assert RP2Decimal(str(busd_transaction_out.crypto_out_with_fee)) == RP2Decimal("200.00")
        assert busd_transaction_out.fiat_out_no_fee is None
        assert busd_transaction_out.fiat_fee is None

        # Buy BTC autoinvest
        assert busd_transaction_in.asset == "BUSD"
        assert busd_transaction_in.timestamp == "2022-02-27 12:11:00+0000"
        assert busd_transaction_in.transaction_type == Keyword.BUY.value.capitalize()
        assert busd_transaction_in.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(busd_transaction_in.crypto_in) == RP2Decimal("110.00")
        assert RP2Decimal(str(busd_transaction_in.crypto_fee)) == RP2Decimal("0.055")
        assert busd_transaction_in.fiat_in_no_fee is None
        assert busd_transaction_in.fiat_in_with_fee is None
        assert busd_transaction_in.fiat_fee is None

        assert eth_transaction_out.asset == "ETH"
        assert eth_transaction_out.timestamp == "2022-02-27 12:11:00+0000"
        assert eth_transaction_out.transaction_type == Keyword.SELL.value.capitalize()
        assert eth_transaction_out.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(eth_transaction_out.crypto_out_no_fee) == RP2Decimal("0.05")
        assert RP2Decimal(eth_transaction_out.crypto_fee) == ZERO
        assert RP2Decimal(str(eth_transaction_out.crypto_out_with_fee)) == RP2Decimal("0.05")
        assert eth_transaction_out.fiat_out_no_fee is None
        assert eth_transaction_out.fiat_fee is None

    def test_transfers(self) -> None:
        plugin = InputPlugin(
            account_holder="tester",
            trades_csv_file=None,
            transfers_csv_file="input/test_pionex_transfers.csv",
            native_fiat="USD",
        )

        result = plugin.load(US())

        # 1 BUSD deposit +
        # 1 BUSD withdrawal = 2
        assert len(result) == 2

        busd_deposit_transaction: IntraTransaction = result[0]  # type: ignore
        busd_withdrawal_transaction: IntraTransaction = result[1]  # type: ignore

        assert busd_deposit_transaction.asset == "BUSD"
        assert busd_deposit_transaction.timestamp == "2022-02-27 12:02:00+0000"
        assert busd_deposit_transaction.unique_id == "12345"
        assert busd_deposit_transaction.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(busd_deposit_transaction.crypto_received) == RP2Decimal("115")
        assert busd_deposit_transaction.crypto_sent == Keyword.UNKNOWN.value
        assert busd_deposit_transaction.from_exchange == Keyword.UNKNOWN.value
        assert busd_deposit_transaction.from_holder == Keyword.UNKNOWN.value
        assert busd_deposit_transaction.to_exchange == "Pionex"
        assert busd_deposit_transaction.to_holder == "tester"

        assert busd_withdrawal_transaction.asset == "BUSD"
        assert busd_withdrawal_transaction.timestamp == "2022-03-25 11:02:00+0000"
        assert busd_withdrawal_transaction.unique_id == "12346"
        assert busd_withdrawal_transaction.spot_price == Keyword.UNKNOWN.value
        assert busd_withdrawal_transaction.crypto_received == Keyword.UNKNOWN.value
        assert RP2Decimal(busd_withdrawal_transaction.crypto_sent) == RP2Decimal("100")
        assert busd_withdrawal_transaction.from_exchange == "Pionex"
        assert busd_withdrawal_transaction.from_holder == "tester"
        assert busd_withdrawal_transaction.to_exchange == Keyword.UNKNOWN.value
        assert busd_withdrawal_transaction.to_holder == Keyword.UNKNOWN.value



# Copyright 2023 Neal Chambers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=protected-access

from datetime import datetime
from os import listdir, makedirs, path, remove, unlink
from typing import Any, List, Optional

from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2RuntimeError

from dali.cache import CACHE_DIR
from dali.historical_bar import HistoricalBar
from dali.plugin.pair_converter.csv.kraken import Kraken

_CACHE_DIRECTORY: str = "output/kraken_test"


class TestKrakenCsvDownload:
    def test_chunking(self, mocker: Any) -> None:
        kraken_csv = Kraken("whatever")

        # Flush test cache directory
        for filename in listdir(_CACHE_DIRECTORY):
            file_path = path.join(_CACHE_DIRECTORY, filename)
            try:
                if path.isfile(file_path):
                    unlink(file_path)
            except RP2RuntimeError:
                pass

        cache_path = path.join(CACHE_DIR, kraken_csv.cache_key())
        if path.exists(cache_path):
            remove(cache_path)

        mocker.patch.object(kraken_csv, "_Kraken__CACHE_DIRECTORY", _CACHE_DIRECTORY)
        if not path.exists(_CACHE_DIRECTORY):
            makedirs(_CACHE_DIRECTORY)
        with open("input/USD_OHLCVT_test.zip", "rb") as file:
            mocker.patch.object(kraken_csv, "_google_file_to_bytes").return_value = file.read()

        test_bar: Optional[HistoricalBar] = kraken_csv.find_historical_bar("USDT", "USD", datetime.fromtimestamp(1601856000))
        files: List[str] = listdir(_CACHE_DIRECTORY)

        # Test if proper price was retrieved and file was chunked
        assert test_bar
        assert test_bar.low == RP2Decimal("1.778")
        assert "USDTUSD_1594080000_5.csv.gz" in files

        test_bar = kraken_csv.find_historical_bar("USDT", "USD", datetime.fromtimestamp(1601683300))

        # Check if price for longer time span was retrieved even though the timestamp doesn't exist in the csv
        # Also that proper price was retrieved from chunked files in the cache folder
        assert test_bar
        assert test_bar.low == RP2Decimal("1.6668")



# Copyright 2022  Steve Davis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# CSV Format:
#   0: "Operation Date"  # For example, "2022-06-05T00:39:14.007Z"
#   1: "Currency Ticker"
#   2: "Operation Type"
#   3: "Operation Amount"
#   4: "Operation Fees"  # This value is often missing or incorrect for receive transactions
#   5: "Operation Hash"
#   6: "Account Name"
#   7: "Account xpub"  # Public key
#   8: "Countervalue Ticker"  # Fiat ticker
#   9: "Countervalue at Operation Date"  # Total fiat value
#  10: "Countervalue at CSV Export"
#
# Note: the Ledger Live software displays this warning message when exporting operation history:
#       The countervalues in the export is provided for information
#       purposes only. Do not rely on such data for accounting, tax,
#       regulation or legal purposes, as they only represent an
#       estimation of the price of the assets at the time of transactions
#       and export, under the valuation methods provided by our
#       service provider, Kaiko


import logging
from csv import reader
from datetime import datetime
from typing import List, Optional
import dateutil

from rp2.abstract_country import AbstractCountry
from rp2.logger import create_logger
from rp2.rp2_decimal import ZERO, RP2Decimal

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.configuration import Keyword
from dali.intra_transaction import IntraTransaction

_SENT: str = "OUT"
_RECV: str = "IN"


class InputPlugin(AbstractInputPlugin):

    __LEDGER: str = "Ledger"

    __TIMESTAMP_INDEX: int = 0
    __CURRENCY_INDEX: int = 1
    __OPERATION_TYPE_INDEX: int = 2
    __QUANTITY_INDEX: int = 3
    __FEE_INDEX: int = 4
    __TRANSACTION_ID_INDEX: int = 5

    __DELIMITER = ","

    __CURRENCY_ALIAS_DICT = {
        "FANTOM": "FTM",
    }

    def __init__(
        self,
        account_holder: str,
        account_nickname: str,
        csv_file: str,
        native_fiat: Optional[str] = None,
    ) -> None:

        super().__init__(account_holder=account_holder, native_fiat=native_fiat)
        self.__account_nickname: str = account_nickname
        self.__csv_file: str = csv_file
        self.__logger: logging.Logger = create_logger(f"{self.__LEDGER}/{self.__account_nickname}/{self.account_holder}")

    def load(self, country: AbstractCountry) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        with open(self.__csv_file, encoding="utf-8") as csv_file:
            lines = reader(csv_file, delimiter=self.__DELIMITER)
            header_found: bool = False
            for line in lines:
                raw_data: str = self.__DELIMITER.join(line)
                if not header_found:
                    # Skip header line
                    header_found = True
                    self.__logger.debug("Header: %s", raw_data)
                    continue
                timestamp: str = line[self.__TIMESTAMP_INDEX]
                timestamp_value: datetime = dateutil.parser.isoparse(timestamp)  # For example, "2022-06-05T00:39:14.007Z"
                self.__logger.debug("Transaction: %s", raw_data)
                currency: str = line[self.__CURRENCY_INDEX]
                currency = self.__CURRENCY_ALIAS_DICT.get(currency, currency)
                transaction_type: str = line[self.__OPERATION_TYPE_INDEX]
                spot_price: str = Keyword.UNKNOWN.value
                crypto_hash: str = line[self.__TRANSACTION_ID_INDEX]
                fee_str: str = line[self.__FEE_INDEX]
                fee_number: RP2Decimal = RP2Decimal(fee_str) if fee_str else ZERO  # Fee is sometimes missing
                quantity_number: RP2Decimal = RP2Decimal(line[self.__QUANTITY_INDEX])

                if quantity_number == ZERO and fee_number > ZERO:
                    self.__logger.warning("Possible dusting attack (fee > 0, total amount = 0): %s", raw_data)
                    continue
                if transaction_type in {_SENT, _RECV}:  # Need example data for sent transactions, untested as of 7/9/2022
                    result.append(
                        IntraTransaction(
                            plugin=self.__LEDGER,
                            unique_id=crypto_hash,
                            raw_data=raw_data,
                            timestamp=f"{timestamp_value}",
                            asset=currency,
                            from_exchange=self.__account_nickname if transaction_type == _SENT else Keyword.UNKNOWN.value,
                            from_holder=self.account_holder if transaction_type == _SENT else Keyword.UNKNOWN.value,
                            to_exchange=self.__account_nickname if transaction_type == _RECV else Keyword.UNKNOWN.value,
                            to_holder=self.account_holder if transaction_type == _RECV else Keyword.UNKNOWN.value,
                            spot_price=spot_price,
                            crypto_sent=str(quantity_number + fee_number) if transaction_type == _SENT else Keyword.UNKNOWN.value,
                            crypto_received=str(quantity_number) if transaction_type == _RECV else Keyword.UNKNOWN.value,
                            notes=None,
                        )
                    )
                else:
                    self.__logger.error("Unsupported transaction type (skipping): %s. Please open an issue at %s", raw_data, self.ISSUES_URL)

        return result



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import unittest
from pathlib import Path
from typing import Dict, List

from dali.abstract_transaction import AbstractTransaction
from dali.cache import CACHE_DIR, load_from_cache, save_to_cache
from dali.in_transaction import InTransaction
from dali.out_transaction import OutTransaction

ROOT_PATH: Path = Path(os.path.dirname(__file__)).parent.absolute()
OUTPUT_PATH: Path = ROOT_PATH / Path("output")


class TestCache(unittest.TestCase):
    def setUp(self) -> None:  # pylint: disable=invalid-name
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_list_cache(self) -> None:
        cache_name: str = "test_list_cache"
        try:
            (ROOT_PATH / CACHE_DIR / cache_name).unlink()
        except FileNotFoundError:
            pass
        transaction_list: List[AbstractTransaction] = [
            InTransaction(
                plugin="my plugin 1",
                unique_id="my unique_id 1",
                raw_data="my raw_data 1",
                timestamp="2021-01-02T08:42:43.882Z",
                asset="BTC",
                exchange="BlockFi",
                holder="Bob",
                transaction_type="inTerest",
                spot_price="1000.0",
                crypto_in="2.0002",
                fiat_fee="0",
                fiat_in_no_fee="2000.2",
                fiat_in_with_fee="2000.2",
            ),
            OutTransaction(
                plugin="my plugin 2",
                unique_id="my unique_id 2",
                raw_data="my raw_data 2",
                timestamp="6/1/2020 3:59:59 -04:00",
                asset="ETH",
                exchange="Coinbase Pro",
                holder="Bob",
                transaction_type="SELL",
                spot_price="900.9",
                crypto_out_no_fee="2.2",
                crypto_fee="0.01",
                notes="my notes 2",
            ),
        ]
        save_to_cache(cache_name, transaction_list)
        loaded_transaction_list: List[AbstractTransaction] = load_from_cache(cache_name)
        self.assertEqual(len(transaction_list), len(loaded_transaction_list))
        for transaction1, transaction2 in zip(transaction_list, loaded_transaction_list):
            self.assertEqual(transaction1, transaction2)
            self.assertEqual(transaction1.__class__, transaction2.__class__)
            self.assertEqual(transaction1.timestamp, transaction2.timestamp)
            self.assertEqual(transaction1.raw_data, transaction2.raw_data)
            if isinstance(transaction1, InTransaction) and isinstance(transaction2, InTransaction):
                self.assertEqual(transaction1.spot_price, transaction2.spot_price)
                self.assertEqual(transaction1.crypto_in, transaction2.crypto_in)
                self.assertEqual(transaction1.crypto_fee, transaction2.crypto_fee)
                self.assertEqual(transaction1.fiat_fee, transaction2.fiat_fee)
            elif isinstance(transaction1, OutTransaction) and isinstance(transaction2, OutTransaction):
                self.assertEqual(transaction1.spot_price, transaction2.spot_price)
                self.assertEqual(transaction1.crypto_out_no_fee, transaction2.crypto_out_no_fee)
                self.assertEqual(transaction1.crypto_fee, transaction2.crypto_fee)
            self.assertEqual(transaction1.notes, transaction2.notes)

    def test_dict_cache(self) -> None:
        cache_name: str = "test_dict_cache"
        try:
            (ROOT_PATH / CACHE_DIR / cache_name).unlink()
        except FileNotFoundError:
            pass
        dictionary: Dict[str, int] = {
            "abc": 12,
            "def": 87,
            "ghi": 98,
        }
        save_to_cache(cache_name, dictionary)
        loaded_dictionary: Dict[str, int] = load_from_cache(cache_name)
        self.assertEqual(dictionary, loaded_dictionary)


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This plugin parses a newer (circa 2023) version of the CSV format generated by Trezor Suite.
# CSV Format: timestamp, date, time, type, transaction_id, fee, fee unit, address, label, amount, amount unit, fiat, other

import logging
from csv import reader
from datetime import datetime
import re
from dateutil.parser import parse
from typing import List, Optional

from rp2.abstract_country import AbstractCountry
from rp2.logger import create_logger
from rp2.rp2_decimal import ZERO, RP2Decimal

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.configuration import Keyword
from dali.intra_transaction import IntraTransaction

_SENT: str = "SENT"
_RECV: str = "RECV"


class InputPlugin(AbstractInputPlugin):

    __TREZOR_V2: str = "Trezor V2"

    __DATE_INDEX: int = 1
    __TIME_INDEX: int = 2
    __TYPE_INDEX: int = 3
    __TRANSACTION_ID_INDEX: int = 4
    __FEE_INDEX: int = 5
    __AMOUNT_INDEX: int = 9

    __DELIMITER = ";"

    def __init__(
        self,
        account_holder: str,
        account_nickname: str,
        currency: str,
        csv_file: str,
        native_fiat: Optional[str] = None,
    ) -> None:

        super().__init__(account_holder=account_holder, native_fiat=native_fiat)
        self.__account_nickname: str = account_nickname
        self.__currency: str = currency
        self.__csv_file: str = csv_file

        self.__logger: logging.Logger = create_logger(f"{self.__TREZOR_V2}/{currency}/{self.__account_nickname}/{self.account_holder}")

    def load(self, country: AbstractCountry) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        with open(self.__csv_file, encoding="utf-8") as csv_file:
            lines = reader(csv_file, delimiter=self.__DELIMITER)
            header_found: bool = False
            for line in lines:
                raw_data: str = self.__DELIMITER.join(line)
                if not header_found:
                    # Skip header line
                    header_found = True
                    self.__logger.debug("Header: %s", raw_data)
                    continue
                timestamp: str = f"{line[self.__DATE_INDEX]} {line[self.__TIME_INDEX]}"
                timestamp = re.sub(r"GMT([+-])([0-9]+)", r"\1\2:00", timestamp)
                timestamp_value: datetime = parse(timestamp)
                self.__logger.debug("Transaction: %s", raw_data)
                transaction_type: str = line[self.__TYPE_INDEX]
                spot_price: str = Keyword.UNKNOWN.value
                crypto_hash: str = line[self.__TRANSACTION_ID_INDEX]
                fee_number: RP2Decimal = RP2Decimal(line[self.__FEE_INDEX])
                amount_number: RP2Decimal = RP2Decimal(line[self.__AMOUNT_INDEX])

                if amount_number == ZERO and fee_number > ZERO:
                    self.__logger.warning("Possible dusting attack (fee > 0, total = 0): %s", raw_data)
                    continue
                if transaction_type in {_RECV, _SENT}:
                    result.append(
                        IntraTransaction(
                            plugin=self.__TREZOR_V2,
                            unique_id=crypto_hash,
                            raw_data=raw_data,
                            timestamp=f"{timestamp_value}",
                            asset=self.__currency,
                            from_exchange=self.__account_nickname if transaction_type == _SENT else Keyword.UNKNOWN.value,
                            from_holder=self.account_holder if transaction_type == _SENT else Keyword.UNKNOWN.value,
                            to_exchange=self.__account_nickname if transaction_type == _RECV else Keyword.UNKNOWN.value,
                            to_holder=self.account_holder if transaction_type == _RECV else Keyword.UNKNOWN.value,
                            spot_price=spot_price,
                            crypto_sent=str(amount_number + fee_number) if transaction_type == _SENT else Keyword.UNKNOWN.value,
                            crypto_received=str(amount_number) if transaction_type == _RECV else Keyword.UNKNOWN.value,
                            notes=None,
                        )
                    )
                else:
                    self.__logger.error("Unsupported transaction type (skipping): %s. Please open an issue at %s", raw_data, self.ISSUES_URL)

        return result



# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rp2.plugin.country.us import US
from rp2.rp2_decimal import RP2Decimal

from dali.configuration import Keyword
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.plugin.input.csv.bitbank_supplemental import InputPlugin


class TestBitbank:
    def test_withdrawals(self) -> None:
        plugin = InputPlugin(
            account_holder="tester",
            withdrawals_csv_file="input/test_bitbank_withdrawals.csv",
            withdrawals_code="XLM",
            deposits_csv_file="input/test_bitbank_deposits.csv",
            deposits_code="JPY",
            native_fiat="USD",
        )

        result = plugin.load(US())

        # 1 XLM withdrawal for 900 XLM
        # 1 JPY deposit for 6900 JPY
        assert len(result) == 2

        xlm_transaction: IntraTransaction = result[0]  # type: ignore
        jpy_transaction: InTransaction = result[1]  # type: ignore

        assert xlm_transaction.asset == "XLM"
        assert xlm_transaction.timestamp == "2022-04-20 07:20:00+0000"
        assert xlm_transaction.unique_id == "TXID12345"
        assert xlm_transaction.spot_price == Keyword.UNKNOWN.value
        assert xlm_transaction.crypto_received == Keyword.UNKNOWN.value
        assert RP2Decimal(xlm_transaction.crypto_sent) == RP2Decimal("900.01")
        assert xlm_transaction.from_exchange == "Bitbank.cc"
        assert xlm_transaction.from_holder == "tester"
        assert xlm_transaction.to_exchange == Keyword.UNKNOWN.value
        assert xlm_transaction.to_holder == Keyword.UNKNOWN.value

        assert jpy_transaction.asset == "JPY"
        assert jpy_transaction.timestamp == "2022-04-19 22:19:00+0000"
        assert RP2Decimal(jpy_transaction.spot_price) == RP2Decimal("1")
        assert RP2Decimal(jpy_transaction.crypto_in) == RP2Decimal("6900")
        assert jpy_transaction.crypto_fee is None
        assert RP2Decimal(str(jpy_transaction.fiat_in_no_fee)) == RP2Decimal("6900")
        assert RP2Decimal(str(jpy_transaction.fiat_in_with_fee)) == RP2Decimal("6900")
        assert jpy_transaction.fiat_ticker == "JPY"



# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Binance.com REST plugin links:
# REST API: https://binance-docs.github.io/apidocs/
# Authentication: https://binance-docs.github.io/apidocs/spot/en/#introduction
# Endpoint: https://api.binance.com

# CCXT documentation:
# https://docs.ccxt.com/en/latest/index.html

# pylint: disable=too-many-lines

import json
import re
from datetime import datetime
from multiprocessing.pool import ThreadPool
from typing import Any, Dict, List, Optional, Union

from ccxt import Exchange, binance

from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2RuntimeError

from dali.abstract_ccxt_input_plugin import (
    AbstractCcxtInputPlugin,
    ProcessOperationResult,
    Trade,
)
from dali.ccxt_pagination import (
    AbstractPaginationDetailSet,
    DateBasedPaginationDetailSet,
)
from dali.configuration import Keyword
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.out_transaction import OutTransaction

# Native format keywords
_ACCOUNT_PROFITS: str = "accountProfits"
_ALGO: str = "algo"
_ALGO_NAME: str = "algoName"
_AMOUNT: str = "amount"
_ASSET: str = "asset"
_BEGIN_TIME: str = "beginTime"
_BUY: str = "buy"  # CCXT only variable
_COIN: str = "coin"
_COIN_NAME: str = "coinName"
_COST: str = "cost"  # CCXT only variable
_CREATE_TIME: str = "createTime"
_CRYPTOCURRENCY: str = "cryptoCurrency"
_CURRENCY: str = "currency"  # CCXT only variable
_DAILY: str = "DAILY"
_DATA: str = "data"
_DATE_TIME: str = "datetime"  # CCXT only variable
_DELIVER_DATE: str = "deliverDate"
_DEPOSIT: str = "deposit"  # CCXT only variable
_DIV_TIME: str = "divTime"
_END_TIME: str = "endTime"
_EN_INFO: str = "enInfo"
_FEE: str = "fee"
_FIAT_CURRENCY: str = "fiatCurrency"
_ID: str = "id"  # CCXT only variable
_INDICATED_AMOUNT: str = "indicatedAmount"
_INFO: str = "info"
_INSERT_TIME: str = "insertTime"
_INTEREST_PARAMETER: str = "INTEREST"
_INTEREST_FIELD: str = "interest"
_IS_DUST: str = "isDust"
_IS_FIAT_PAYMENT: str = "isFiatPayment"
_LEGAL_MONEY: str = "legalMoney"
_LENDING_TYPE: str = "lendingType"
_LIMIT: str = "limit"
_LOCK_PERIOD: str = "lockPeriod"
_OBTAIN_AMOUNT: str = "obtainAmount"
_ORDER: str = "order"  # CCXT only variable
_ORDER_NO: str = "orderNo"
_PAGE_INDEX: str = "pageIndex"
_PAGE_SIZE: str = "pageSize"
_POSITION_ID: str = "positionId"
_PRICE: str = "price"
_PRODUCT: str = "product"
_PROFIT_AMOUNT: str = "profitAmount"
_REDEMPTION: str = "REDEMPTION"
_ROWS: str = "rows"
_SELL: str = "sell"  # CCXT only variable
_SIDE: str = "side"  # CCXT only variable
_SIZE: str = "size"
_STAKING: str = "STAKING"
_START_TIME: str = "startTime"
_STATUS: str = "status"
_SOURCE_AMOUNT: str = "sourceAmount"
_SUBSCRIPTION: str = "SUBSCRIPTION"
_SYMBOL: str = "symbol"
_TIME: str = "time"
_TIMESTAMP: str = "timestamp"  # CCXT only variable
_TRAN_ID: str = "tranId"
_TRANSACTION_TYPE: str = "transactionType"
_TOTAL: str = "total"
_TOTAL_FEE: str = "totalFee"
_TOTAL_NUM: str = "totalNum"
_TYPE: str = "type"
_TX_ID: str = "txid"  # CCXT doesn't capitalize I
_TXN_TYPE: str = "txnType"
_UPDATE_TIME: str = "updateTime"
_USERNAME: str = "userName"
_WITHDRAWAL: str = "withdrawal"  # CCXT only variable

# Time period constants
_NINETY_DAYS_IN_MS: int = 7776000000
_THIRTY_DAYS_IN_MS: int = 2592000000
_ONE_DAY_IN_MS: int = 86400000
_MS_IN_SECOND: int = 1000

# Record limits
_DEPOSIT_RECORD_LIMIT: int = 1000
_DIVIDEND_RECORD_LIMIT: int = 500
_DUST_TRADE_RECORD_LIMIT: int = 100
_INTEREST_SIZE_LIMIT: int = 100
_MINING_PAGE_LIMIT: int = 200
_TRADE_RECORD_LIMIT: int = 1000
_WITHDRAWAL_RECORD_LIMIT: int = 1000

# Types of Binance Dividends
_BNB_VAULT = "BNB Vault"
_ETH_STAKING = "ETH 2.0 Staking"
_FLEXIBLE = "Flexible"
_FLEXIBLE_SAVINGS = "Flexible Savings"
_LAUNCH_POOL = "Launchpool"
_LOCKED = "Locked"
_LOCKED_SAVINGS = "Locked Savings"
_LOCKED_STAKING = "Locked Staking"
_SOLO_AIRDROP = "SOLO airdrop"
_GENERAL_STAKING = "STAKING"

_AIRDROP_LIST = [_SOLO_AIRDROP]
_INTEREST_LIST = [_FLEXIBLE, _FLEXIBLE_SAVINGS, _LOCKED, _LOCKED_SAVINGS]
_STAKING_LIST = [_ETH_STAKING, _LOCKED_STAKING, _BNB_VAULT, _LAUNCH_POOL, _GENERAL_STAKING]


class InputPlugin(AbstractCcxtInputPlugin):

    __EXCHANGE_NAME: str = "Binance.com"
    __PLUGIN_NAME: str = "Binance.com_REST"
    __DEFAULT_THREAD_COUNT: int = 1

    def __init__(
        self,
        account_holder: str,
        api_key: str,
        api_secret: str,
        native_fiat: str,
        username: Optional[str] = None,
        thread_count: Optional[int] = __DEFAULT_THREAD_COUNT,
    ) -> None:

        self.__api_key = api_key
        self.__api_secret = api_secret
        # We will have a default start time of July 13th, 2017 since Binance Exchange officially launched on July 14th Beijing Time.
        super().__init__(account_holder, datetime(2017, 7, 13, 0, 0, 0, 0), native_fiat, thread_count)
        self.__username = username

        # We have to know what markets and algos are on Binance so that we can pull orders using the market
        self.__algos: List[str] = []

    def exchange_name(self) -> str:
        return self.__EXCHANGE_NAME

    def plugin_name(self) -> str:
        return self.__PLUGIN_NAME

    def _initialize_client(self) -> binance:
        return binance(
            {
                "apiKey": self.__api_key,
                "enableRateLimit": True,
                "secret": self.__api_secret,
            }
        )

    @property
    def _client(self) -> binance:
        super_client: Exchange = super()._client
        if not isinstance(super_client, binance):
            raise RP2RuntimeError("Exchange is not instance of class binance.")
        return super_client

    def _get_algos(self) -> List[str]:
        if self.__algos:
            return self.__algos
        if self.__username:
            binance_algos = self._client.sapiGetMiningPubAlgoList()
            for algo in binance_algos[_DATA]:
                self._logger.debug("Algo: %s", json.dumps(algo))
                self.__algos.append(algo[_ALGO_NAME])
            return self.__algos
        return []

    def _get_process_deposits_pagination_detail_set(self) -> Optional[AbstractPaginationDetailSet]:
        return DateBasedPaginationDetailSet(
            limit=_DEPOSIT_RECORD_LIMIT,
            exchange_start_time=self._start_time_ms,
            window=_NINETY_DAYS_IN_MS,
        )

    def _get_process_withdrawals_pagination_detail_set(self) -> Optional[AbstractPaginationDetailSet]:
        return DateBasedPaginationDetailSet(
            limit=_WITHDRAWAL_RECORD_LIMIT,
            exchange_start_time=self._start_time_ms,
            window=_NINETY_DAYS_IN_MS,
        )

    def _get_process_trades_pagination_detail_set(self) -> Optional[AbstractPaginationDetailSet]:
        return DateBasedPaginationDetailSet(
            limit=_TRADE_RECORD_LIMIT,
            exchange_start_time=self._start_time_ms,
            markets=self._get_markets(),
        )

    ### Multiple transaction processing

    def _process_gains(
        self,
        in_transactions: List[InTransaction],
        out_transactions: List[OutTransaction],
    ) -> None:

        ### Regular Dividends from Staking (including Eth staking) and Savings (Lending) after around May 8th, 2021 01:00 UTC.

        # We need milliseconds for Binance
        current_start = self._start_time_ms
        now_time = int(datetime.now().timestamp()) * _MS_IN_SECOND
        processing_result_list: List[Optional[ProcessOperationResult]] = []

        # The exact moment when Binance switched to unified dividends is unknown/unpublished.
        # This allows us an educated guess.
        earliest_record_epoch: int = 0

        # We will pull in 30 day periods. This allows for 16 assets with daily dividends.
        current_end = current_start + _THIRTY_DAYS_IN_MS

        while current_start < now_time:
            self._logger.debug("Pulling dividends/subscriptions/redemptions from %s to %s", current_start, current_end)

            # CCXT doesn't have a standard way to pull income, we must use the underlying API endpoint
            dividends = self._client.sapiGetAssetAssetDividend(params=({_START_TIME: current_start, _END_TIME: current_end, _LIMIT: _DIVIDEND_RECORD_LIMIT}))
            # {
            #     "rows":[
            #         {
            #             "id":1637366104,
            #             "amount":"10.00000000",
            #             "asset":"BHFT",
            #             "divTime":1563189166000,
            #             "enInfo":"BHFT distribution",
            #             "tranId":2968885920
            #         },
            #         {
            #             "id":1631750237,
            #             "amount":"10.00000000",
            #             "asset":"BHFT",
            #             "divTime":1563189165000,
            #             "enInfo":"BHFT distribution",
            #             "tranId":2968885920
            #         }
            #     ],
            #     "total":2
            # }

            self._logger.debug("Pulled a total of %s records for %s to %s", dividends[_TOTAL], current_start, current_end)

            # If user received more than 500 dividends in a 30 day period we need to shrink the window.
            if int(dividends[_TOTAL]) <= _DIVIDEND_RECORD_LIMIT:
                current_start = current_end + 1
                current_end = current_start + _THIRTY_DAYS_IN_MS
                with ThreadPool(self._thread_count) as pool:
                    processing_result_list = pool.map(self._process_dividend, dividends[_ROWS])

                for processing_result in processing_result_list:
                    if processing_result is None:
                        continue
                    if processing_result.in_transactions:
                        in_transactions.extend(processing_result.in_transactions)
            else:
                # Using implicit API so we need to follow Binance order, which sends latest record first ([0])
                # CCXT standard API sorts by timestamp, so latest record is last ([499])
                number_of_excess_records = int(dividends[_TOTAL]) - _DIVIDEND_RECORD_LIMIT
                current_end = int(dividends[_ROWS][number_of_excess_records][_DIV_TIME])  # times are inclusive
                self._logger.debug("Readjusting time window end to %s from %s", current_end, current_start + _THIRTY_DAYS_IN_MS)
                # current_end = current_start + _THIRTY_DAYS_IN_MS

            if not earliest_record_epoch and int(dividends[_TOTAL]) > 0:
                earliest_record_epoch = int(dividends[_ROWS][-1][_DIV_TIME]) - 1

            # We need to track subscription and redemption amounts since Binance will take a fee equal to the amount of interest
            # earned during the lock period if the user prematurely redeems their funds.

        # Old system Locked Savings

        old_savings: bool = False

        # Reset window
        current_start = self._start_time_ms
        current_end = current_start + _THIRTY_DAYS_IN_MS

        # The cummulative interest from a positionID
        total_current_interest: Dict[int, RP2Decimal] = {}

        # The cummulative interest payments made to each positionID
        total_current_payments: Dict[int, int] = {}

        # Subscriptions organized [asset][amount] = timestamp in milliseconds
        current_subscriptions: Dict[str, Dict[str, Dict[str, str]]] = {}

        # We will step backward in time from the switch over
        while current_start < now_time:

            self._logger.debug("Pulling locked staking from older api system from %s to %s", current_start, current_end)

            locked_staking = self._client.sapi_get_staking_stakingrecord(
                params=({_START_TIME: current_start, _END_TIME: current_end, _PRODUCT: _STAKING, _TXN_TYPE: _INTEREST_PARAMETER, _SIZE: _INTEREST_SIZE_LIMIT})
            )
            # [
            #   {
            #       'positionId': '7146912',
            #       'time': '1624233772000',
            #       'asset': 'BTC',
            #       'amount': '0.017666',
            #       'status': 'SUCCESS'
            #   },
            #   {
            #       'positionId': '7147052',
            #       'time': '1624147893000',
            #       'asset': 'BTC',
            #       'amount': '0.0176665',
            #       'status': 'SUCCESS'
            #   }
            # ]
            # NOTE: All values are str
            processing_result_list = []
            for stake_dividend in locked_staking:
                if int(stake_dividend[_TIME]) < earliest_record_epoch:
                    (self._logger).debug("Locked Staking (OLD): %s", json.dumps(stake_dividend))
                    stake_dividend[_EN_INFO] = "Locked Staking/Savings (OLD)"
                    stake_dividend[_ID] = Keyword.UNKNOWN.value
                    stake_dividend[_DIV_TIME] = stake_dividend[_TIME]
                    processing_result_list.append(self._process_gain(stake_dividend, Keyword.STAKING))
                    old_savings = True

                # Early redemption penalty tracking. Needs to be recorded even for new system.
                position_id: int = stake_dividend[_POSITION_ID]
                total_current_interest[position_id] = total_current_interest.get(position_id, ZERO) + RP2Decimal(str(stake_dividend[_AMOUNT]))
                total_current_payments[position_id] = total_current_payments.get(position_id, 0) + 1

            for processing_result in processing_result_list:
                if processing_result is None:
                    continue
                if processing_result.in_transactions:
                    in_transactions.extend(processing_result.in_transactions)

            locked_subscriptions = self._client.sapi_get_staking_stakingrecord(
                params=({_START_TIME: current_start, _END_TIME: current_end, _PRODUCT: _STAKING, _TXN_TYPE: _SUBSCRIPTION, _SIZE: _INTEREST_SIZE_LIMIT})
            )
            # [
            #     {
            #         "time": "1624147893000",
            #         "asset": "BTC",
            #         "amount": "1",
            #         "lockPeriod": "10",
            #         "type": "NORMAL",
            #         "status": "SUCCESS",
            #     },
            #     {
            #         "time": "1624147893000",
            #         "asset": "BTC",
            #         "amount": "1",
            #         "lockPeriod": "10",
            #         "type": "NORMAL",
            #         "status": "SUCCESS",
            #     }
            # ]
            # NOTE: all values are str

            for subscription in locked_subscriptions:

                # If the dict already exists add another key, if not add new dict
                if current_subscriptions.get(subscription[_ASSET]):
                    current_subscriptions[subscription[_ASSET]][f"{RP2Decimal(subscription[_AMOUNT]):.13f}"] = subscription
                else:
                    current_subscriptions[subscription[_ASSET]] = {f"{RP2Decimal(subscription[_AMOUNT]):.13f}": subscription}

            locked_redemptions = self._client.sapi_get_staking_stakingrecord(
                params=({_START_TIME: current_start, _END_TIME: current_end, _PRODUCT: _STAKING, _TXN_TYPE: _REDEMPTION, _SIZE: _INTEREST_SIZE_LIMIT})
            )
            # [
            #         {
            #             "positionId": "12345",
            #             "time": "1624147893000"
            #             "asset": "BTC",
            #             "amount": "1",
            #             "deliverDate": "1624147895000"
            #             "status": "PAID",
            #         },
            #         {
            #             "positionId": "12346",
            #             "time": "1624147993000"
            #             "asset": "BTC",
            #             "amount": "0.95",
            #             "deliverDate": "1624148093000"
            #             "status": "PAID",
            #         }
            # ]
            # NOTE: all values are str

            for redemption in locked_redemptions:

                redemption_amount: str = f"{RP2Decimal(redemption[_AMOUNT]):.13f}"

                # Check if there is a subscription with this asset and if the redemption amount is equal to the subscription amount
                if redemption[_ASSET] in current_subscriptions and redemption_amount not in current_subscriptions[redemption[_ASSET]]:

                    # If they do not equal we need to calculate what the amended principal should be based on total interest paid to that productId
                    total_interest_earned: RP2Decimal = total_current_interest[redemption[_POSITION_ID]]
                    original_principal: str = f"{(RP2Decimal(redemption[_AMOUNT]) + RP2Decimal(str(total_interest_earned))):.13f}"
                    earliest_redemption_timestamp: int = 0

                    if str(original_principal) in current_subscriptions[redemption[_ASSET]]:

                        subscription_time: int = int(current_subscriptions[redemption[_ASSET]][str(original_principal)][_TIME])
                        lockperiod_in_ms: int = int(current_subscriptions[redemption[_ASSET]][str(original_principal)][_LOCK_PERIOD]) * _ONE_DAY_IN_MS
                        earliest_redemption_timestamp = subscription_time + lockperiod_in_ms

                    else:
                        raise RP2RuntimeError(
                            f"Internal Error: Principal ({original_principal}) minus paid interest ({RP2Decimal(str(total_interest_earned))}) does not equal"
                            f" returned principal ({RP2Decimal(redemption[_AMOUNT])}) on locked savings position ID - {redemption[_POSITION_ID]}."
                        )

                    # There is some lag time between application for the subscription and when the subscription actually starts ~ 2 days
                    if (int(redemption[_TIME]) - int(earliest_redemption_timestamp)) < 2 * _ONE_DAY_IN_MS:
                        out_transactions.append(
                            OutTransaction(
                                plugin=self.__PLUGIN_NAME,
                                unique_id=Keyword.UNKNOWN.value,
                                raw_data=json.dumps(redemption),
                                timestamp=self._rp2_timestamp_from_ms_epoch(redemption[_DELIVER_DATE]),
                                asset=redemption[_ASSET],
                                exchange=self.__EXCHANGE_NAME,
                                holder=self.account_holder,
                                transaction_type=Keyword.FEE.value,
                                spot_price=Keyword.UNKNOWN.value,
                                crypto_out_no_fee="0",
                                crypto_fee=str(total_interest_earned),
                                crypto_out_with_fee=str(total_interest_earned),
                                fiat_out_no_fee=None,
                                fiat_fee=None,
                                notes=(f"Penalty Fee for {redemption[_POSITION_ID]}"),
                            )
                        )

                    else:
                        raise RP2RuntimeError(
                            f"Internal Error: The redemption time ({self._rp2_timestamp_from_ms_epoch(redemption[_TIME])}) is not in the redemption window "
                            f"({self._rp2_timestamp_from_ms_epoch(str(earliest_redemption_timestamp))} + 2 days)."
                        )

                elif redemption[_ASSET] in current_subscriptions and redemption_amount in current_subscriptions[redemption[_ASSET]]:

                    self._logger.debug("Locked Savings positionId %s redeemed successfully.", redemption[_POSITION_ID])

                else:
                    raise RP2RuntimeError(f"Internal Error: Orphaned Redemption. Please open an issue at {self.ISSUES_URL}.")

            # if we returned the limit, we need to roll the window forward to the last time
            if len(locked_redemptions) < _INTEREST_SIZE_LIMIT:
                current_start = current_end + 1
                current_end = current_start + _THIRTY_DAYS_IN_MS
            else:
                current_start = now_time - 1  # int(locked_redemptions[0][_TIME]) + 1
                current_end = now_time  # current_start + _THIRTY_DAYS_IN_MS

        # Old system Flexible Savings

        # Reset window
        current_start = self._start_time_ms
        current_end = current_start + _THIRTY_DAYS_IN_MS

        # We will step backward in time from the switch over
        while current_start < earliest_record_epoch:

            self._logger.debug("Pulling flexible saving from older api system from %s to %s", current_start, current_end)

            flexible_saving = self._client.sapi_get_lending_union_interesthistory(
                params=({_START_TIME: current_start, _END_TIME: current_end, _LENDING_TYPE: _DAILY, _SIZE: _INTEREST_SIZE_LIMIT})
            )
            # [
            #     {
            #         "asset": "BUSD",
            #         "interest": "0.00006408",
            #         "lendingType": "DAILY",
            #         "productName": "BUSD",
            #         "time": 1577233578000
            #     },
            #     {
            #         "asset": "USDT",
            #         "interest": "0.00687654",
            #         "lendingType": "DAILY",
            #         "productName": "USDT",
            #         "time": 1577233562000
            #     }
            # ]
            processing_result_list = []
            for saving in flexible_saving:
                self._logger.debug("Flexible Saving: %s", json.dumps(saving))
                saving[_EN_INFO] = "Flexible Savings (OLD)"
                saving[_ID] = Keyword.UNKNOWN.value
                saving[_DIV_TIME] = saving[_TIME]
                saving[_AMOUNT] = saving[_INTEREST_FIELD]
                processing_result_list.append(self._process_gain(saving, Keyword.INTEREST))
                old_savings = True

            for processing_result in processing_result_list:
                if processing_result is None:
                    continue
                if processing_result.in_transactions:
                    in_transactions.extend(processing_result.in_transactions)

            # if we returned the limit, we need to roll the window forward to the last time
            if len(flexible_saving) < _INTEREST_SIZE_LIMIT:
                current_start = current_end + 1
                current_end = current_start + _THIRTY_DAYS_IN_MS
            else:
                current_start = int(flexible_saving[0][_TIME]) + 1
                current_end = current_start + _THIRTY_DAYS_IN_MS

            current_end = min(current_end, earliest_record_epoch)

        if old_savings:
            # Since we are making a guess at the cut off, there might be errors.
            self._logger.warning(
                "Pre-May 8th, 2021 savings detected. Please be aware that there may be duplicate or missing savings records around May 8th, 2021."
            )

        ### Mining Income

        # username is only required when pulling mining data
        for algo in self._get_algos():
            # Binance uses pages for mining payments
            current_page = 1
            while True:
                results = self._client.sapiGetMiningPaymentList(
                    params=({_ALGO: algo, _USERNAME: self.__username, _PAGE_INDEX: current_page, _PAGE_SIZE: _MINING_PAGE_LIMIT})
                )
                # {
                #   "code": 0,
                #   "msg": "",
                #   "data": {
                #     "accountProfits": [
                #       {
                #         "time": 1586188800000,            // Mining date
                #         "type": "31", // 0:Mining Wallet,5:Mining Address,7:Pool Savings,
                #           8:Transferred,31:Income Transfer ,32:Hashrate Resale-Mining Wallet 33:Hashrate Resale-Pool Savings
                #         "hashTransfer": null,            // Transferred Hashrate
                #         "transferAmount": null,          // Transferred Income
                #         "dayHashRate": 129129903378244,  // Daily Hashrate
                #         "profitAmount": 8.6083060304,   //Earnings Amount
                #         "coinName":"BTC",              // Coin Type
                #         "status": "2"    //Status：0:Unpaid， 1:Paying  2：Paid
                #       },
                #       {
                #         "time": 1607529600000,
                #         "coinName": "BTC",
                #         "type": "0", // String
                #         "dayHashRate": 9942053925926,
                #         "profitAmount": 0.85426469,
                #         "hashTransfer": 200000000000,
                #         "transferAmount": 0.02180958,
                #         "status": "2"
                #       },
                #       {
                #         "time": 1607443200000,
                #         "coinName": "BTC",
                #         "type": "31",
                #         "dayHashRate": 200000000000,
                #         "profitAmount": 0.02905916,
                #         "hashTransfer": null,
                #         "transferAmount": null,
                #         "status": "2"
                #       }
                #     ],
                #     "totalNum": 3,          // Total Rows
                #     "pageSize": 20          // Rows per page
                #   }
                # }

                if results[_DATA][_TOTAL_NUM] != "0":
                    profits: List[Dict[str, Union[int, str]]] = results[_DATA][_ACCOUNT_PROFITS]
                    processing_result_list = []
                    for result in profits:
                        self._logger.debug("Mining profit: %s", json.dumps(result))

                        # Currently the plugin only supports standard mining deposits
                        # Payment must also be made (status=2) in order to be counted
                        if result[_TYPE] == "0" and result[_STATUS] == "2":
                            processing_result_list.append(self._process_gain(result, Keyword.MINING))
                        else:
                            self._logger.error(
                                "WARNING: Unsupported Mining Transaction Type: %s.\nFull Details: %s\nPlease open an issue at %s.",
                                result[_TYPE],
                                json.dumps(result),
                                self.ISSUES_URL,
                            )

                    for processing_result in processing_result_list:
                        if processing_result is None:
                            continue
                        if processing_result.in_transactions:
                            in_transactions.extend(processing_result.in_transactions)

                    if len(profits) == _MINING_PAGE_LIMIT:
                        current_page += 1
                    else:
                        break
                else:
                    break

    def _process_implicit_api(  # pylint: disable=unused-argument
        self,
        in_transactions: List[InTransaction],
        out_transactions: List[OutTransaction],
        intra_transactions: List[IntraTransaction],
    ) -> None:

        # We need milliseconds for Binance
        now_time = int(datetime.now().timestamp()) * _MS_IN_SECOND

        processing_result_list: List[Optional[ProcessOperationResult]] = []

        # Crypto Bought with fiat. Technically this is a deposit of fiat that is used for a market order that fills immediately.
        # No limit on the date range
        # fiat payments takes the 'beginTime' param in contrast to other functions that take 'startTime'
        fiat_payments = self._client.sapiGetFiatPayments(params=({_TRANSACTION_TYPE: 0, _BEGIN_TIME: self._start_time_ms, _END_TIME: now_time}))
        # {
        #   "code": "000000",
        #   "message": "success",
        #   "data": [
        #   {
        #      "orderNo": "353fca443f06466db0c4dc89f94f027a",
        #      "sourceAmount": "20.0",  // Fiat trade amount
        #      "fiatCurrency": "EUR",   // Fiat token
        #      "obtainAmount": "4.462", // Crypto trade amount
        #      "cryptoCurrency": "LUNA",  // Crypto token
        #      "totalFee": "0.2",    // Trade fee
        #      "price": "4.437472",
        #      "status": "Failed",  // Processing, Completed, Failed, Refunded
        #      "createTime": 1624529919000,
        #      "updateTime": 1624529919000
        #   }
        #   ],
        #   "total": 1,
        #   "success": true
        # }
        if _DATA in fiat_payments:
            with ThreadPool(self._thread_count) as pool:
                processing_result_list = pool.map(self._process_fiat_payment, fiat_payments[_DATA])

            for processing_result in processing_result_list:
                if processing_result is None:
                    continue
                if processing_result.in_transactions:
                    in_transactions.extend(processing_result.in_transactions)
                if processing_result.out_transactions:
                    out_transactions.extend(processing_result.out_transactions)

        # Process actual fiat deposits (no limit on the date range)
        # Fiat deposits can also be pulled via CCXT fetch_deposits by cycling through legal_money
        # Using the underlying api endpoint is faster for Binance.
        # Note that this is the same endpoint as withdrawls, but with _TRANSACTION_TYPE set to 0 (for deposits)
        fiat_deposits = self._client.sapiGetFiatOrders(params=({_TRANSACTION_TYPE: 0, _START_TIME: self._start_time_ms, _END_TIME: now_time}))
        #    {
        #      "code": "000000",
        #      "message": "success",
        #      "data": [
        #        {
        #          "orderNo": "25ced37075c1470ba8939d0df2316e23",
        #          "fiatCurrency": "EUR",
        #          "indicatedAmount": "15.00",
        #          "amount": "15.00",
        #          "totalFee": "0.00",
        #          "method": "card",
        #          "status": "Failed",
        #          "createTime": 1627501026000,
        #          "updateTime": 1627501027000
        #        }
        #      ],
        #      "total": 1,
        #      "success": True
        #    }
        if _DATA in fiat_deposits:
            with ThreadPool(self._thread_count) as pool:
                processing_result_list = pool.map(self._process_fiat_deposit_order, fiat_deposits[_DATA])

            for processing_result in processing_result_list:
                if processing_result is None:
                    continue
                if processing_result.in_transactions:
                    in_transactions.extend(processing_result.in_transactions)

        # Process actual fiat withdrawls (no limit on the date range)
        # Fiat deposits can also be pulled via CCXT fetch_withdrawls by cycling through legal_money
        # Using the underlying api endpoint is faster for Binance.
        # Note that this is the same endpoint as deposits, but with _TRANSACTION_TYPE set to 1 (for withdrawls)
        fiat_withdrawals = self._client.sapiGetFiatOrders(params=({_TRANSACTION_TYPE: 1, _START_TIME: self._start_time_ms, _END_TIME: now_time}))
        #    {
        #      "code": "000000",
        #      "message": "success",
        #      "data": [
        #        {
        #          "orderNo": "25ced37075c1470ba8939d0df2316e23",
        #          "fiatCurrency": "EUR",
        #          "indicatedAmount": "15.00",
        #          "amount": "15.00",
        #          "totalFee": "0.00",
        #          "method": "card",
        #          "status": "Failed",
        #          "createTime": 1627501026000,
        #          "updateTime": 1627501027000
        #        }
        #      ],
        #      "total": 1,
        #      "success": True
        #    }
        if _DATA in fiat_withdrawals:
            with ThreadPool(self._thread_count) as pool:
                processing_result_list = pool.map(self._process_fiat_withdrawal_order, fiat_withdrawals[_DATA])

            for processing_result in processing_result_list:
                if processing_result is None:
                    continue
                if processing_result.out_transactions:
                    out_transactions.extend(processing_result.out_transactions)

        ### Dust Trades

        # We need milliseconds for Binance
        current_start = self._start_time_ms

        # We will pull in 30 day periods
        # If the user has more than 100 dust trades in a 30 day period this will break.
        # Maybe we can set a smaller window in the .ini file?
        current_end = current_start + _THIRTY_DAYS_IN_MS
        while current_start < now_time:
            dust_trades = self._client.fetch_my_dust_trades(params=({_START_TIME: current_start, _END_TIME: current_end}))
            # CCXT returns the same json as .fetch_trades()

            # Binance only returns 100 dust trades per call. If we hit the limit we will have to crawl
            # over each 'dribblet'. Each dribblet can have multiple assets converted into BNB at the same time.
            # If the user converts more than 100 assets at one time, we can not retrieve accurate records.
            if len(dust_trades) == _DUST_TRADE_RECORD_LIMIT:

                first_dribblet = [x for x in dust_trades if x[_DIV_TIME] == dust_trades[0][_DIV_TIME]]
                if len(first_dribblet) == _DUST_TRADE_RECORD_LIMIT:
                    raise RP2RuntimeError(
                        f"Internal error: too many assets dusted at the same time: " f"{self._rp2_timestamp_from_ms_epoch(str(dust_trades[0][_DIV_TIME]))}"
                    )

                with ThreadPool(self._thread_count) as pool:
                    processing_result_list = pool.map(self._process_dust_trade, first_dribblet)

                for processing_result in processing_result_list:
                    if processing_result is None:
                        continue
                    if processing_result.in_transactions:
                        in_transactions.extend(processing_result.in_transactions)
                    if processing_result.out_transactions:
                        out_transactions.extend(processing_result.out_transactions)

                # Shift the call window forward past this dribblet
                current_start = first_dribblet[len(first_dribblet) - 1][_DIV_TIME] + 1
                current_end = current_start + _THIRTY_DAYS_IN_MS
                break

            with ThreadPool(self._thread_count) as pool:
                processing_result_list = pool.map(self._process_dust_trade, dust_trades)

            for processing_result in processing_result_list:
                if processing_result is None:
                    continue
                if processing_result.in_transactions:
                    in_transactions.extend(processing_result.in_transactions)
                if processing_result.out_transactions:
                    out_transactions.extend(processing_result.out_transactions)

            current_start = current_end + 1
            current_end = current_start + _THIRTY_DAYS_IN_MS

    def _process_dividend(self, dividend: Any, notes: Optional[str] = None) -> ProcessOperationResult:
        self._logger.debug("Dividend: %s", json.dumps(dividend))
        if dividend[_EN_INFO] in _STAKING_LIST or re.search("[dD]istribution", dividend[_EN_INFO]) or re.search("staking", dividend[_EN_INFO]):
            return self._process_gain(dividend, Keyword.STAKING, notes)
        if dividend[_EN_INFO] in _INTEREST_LIST:
            return self._process_gain(dividend, Keyword.INTEREST, notes)
        if dividend[_EN_INFO] in _AIRDROP_LIST or re.search("[aA]irdrop", dividend[_EN_INFO]):
            return self._process_gain(dividend, Keyword.AIRDROP, notes)
        self._logger.error("WARNING: Unrecognized Dividend: %s. Please open an issue at %s", dividend[_EN_INFO], self.ISSUES_URL)
        return ProcessOperationResult(in_transactions=[], out_transactions=[], intra_transactions=[])

    def _process_dust_trade(self, dust: Any, notes: Optional[str] = None) -> ProcessOperationResult:
        self._logger.debug("Dust: %s", json.dumps(dust))
        # dust trades have a null id, and if multiple assets are dusted at the same time, all are assigned same ID
        dust_trade: Trade = self._to_trade(dust[_SYMBOL], str(dust[_AMOUNT]), str(dust[_COST]))
        dust[_ID] = f"{dust[_ORDER]}{dust_trade.base_asset}"
        return self._process_buy_and_sell(dust, notes)

    def _process_gain(self, transaction: Any, transaction_type: Keyword, notes: Optional[str] = None) -> ProcessOperationResult:
        self._logger.debug("Gain: %s", json.dumps(transaction))
        in_transaction_list: List[InTransaction] = []

        if transaction_type == Keyword.MINING:
            amount: RP2Decimal = RP2Decimal(str(transaction[_PROFIT_AMOUNT]))
            notes = f"{notes + '; ' if notes else ''}'Mining profit'"
            in_transaction_list.append(
                InTransaction(
                    plugin=self.__PLUGIN_NAME,
                    unique_id=Keyword.UNKNOWN.value,
                    raw_data=json.dumps(transaction),
                    timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_TIME]),
                    asset=transaction[_COIN_NAME],
                    exchange=self.__EXCHANGE_NAME,
                    holder=self.account_holder,
                    transaction_type=transaction_type.value,
                    spot_price=Keyword.UNKNOWN.value,
                    crypto_in=str(amount),
                    crypto_fee=None,
                    fiat_in_no_fee=None,
                    fiat_in_with_fee=None,
                    fiat_fee=None,
                    notes=notes,
                )
            )
        elif RP2Decimal(transaction[_AMOUNT]) != ZERO:  # Sometimes Binance reports interest payments with zero amounts
            amount = RP2Decimal(transaction[_AMOUNT])
            notes = f"{notes + '; ' if notes else ''}{transaction[_EN_INFO]}"

            in_transaction_list.append(
                InTransaction(
                    plugin=self.__PLUGIN_NAME,
                    unique_id=str(transaction[_ID]),  # Binance sometimes has two ids for one tranid
                    raw_data=json.dumps(transaction),
                    timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_DIV_TIME]),
                    asset=transaction[_ASSET],
                    exchange=self.__EXCHANGE_NAME,
                    holder=self.account_holder,
                    transaction_type=transaction_type.value,
                    spot_price=Keyword.UNKNOWN.value,
                    crypto_in=str(amount),
                    crypto_fee=None,
                    fiat_in_no_fee=None,
                    fiat_in_with_fee=None,
                    fiat_fee=None,
                    notes=notes,
                )
            )

        return ProcessOperationResult(in_transactions=in_transaction_list, out_transactions=[], intra_transactions=[])

    def _process_fiat_deposit_order(self, transaction: Any, notes: Optional[str] = None) -> ProcessOperationResult:
        return self._process_fiat_order(transaction, Keyword.BUY, notes)

    def _process_fiat_withdrawal_order(self, transaction: Any, notes: Optional[str] = None) -> ProcessOperationResult:
        return self._process_fiat_order(transaction, Keyword.SELL, notes)

    def _process_fiat_order(self, transaction: Any, transaction_type: Keyword, notes: Optional[str] = None) -> ProcessOperationResult:
        self._logger.debug("Fiat Order (%s): %s", transaction_type.value, json.dumps(transaction))
        in_transaction_list: List[InTransaction] = []
        out_transaction_list: List[OutTransaction] = []
        if transaction[_STATUS] == "Completed":
            amount: RP2Decimal = RP2Decimal(transaction[_INDICATED_AMOUNT])
            fee: RP2Decimal = RP2Decimal(transaction[_TOTAL_FEE])
            notes = f"{notes + '; ' if notes else ''}Fiat {transaction_type.value} of {transaction[_FIAT_CURRENCY]}"
            if transaction_type == Keyword.BUY:
                in_transaction_list.append(
                    InTransaction(
                        plugin=self.__PLUGIN_NAME,
                        unique_id=transaction[_ORDER_NO],
                        raw_data=json.dumps(transaction),
                        timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_CREATE_TIME]),
                        asset=transaction[_FIAT_CURRENCY],
                        exchange=self.__EXCHANGE_NAME,
                        holder=self.account_holder,
                        transaction_type=transaction_type.value,
                        spot_price="1",
                        crypto_in=str(amount),
                        crypto_fee=str(fee),
                        fiat_in_no_fee=None,
                        fiat_in_with_fee=None,
                        fiat_fee=None,
                        fiat_ticker=transaction[_FIAT_CURRENCY],
                        notes=notes,
                    )
                )
            elif transaction_type == Keyword.SELL:
                out_transaction_list.append(
                    OutTransaction(
                        plugin=self.__PLUGIN_NAME,
                        unique_id=transaction[_ORDER_NO],
                        raw_data=json.dumps(transaction),
                        timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_CREATE_TIME]),
                        asset=transaction[_FIAT_CURRENCY],
                        exchange=self.__EXCHANGE_NAME,
                        holder=self.account_holder,
                        transaction_type=Keyword.SELL.value,
                        spot_price="1",
                        crypto_out_no_fee=str(amount),
                        crypto_fee=str(fee),
                        fiat_out_no_fee=None,
                        fiat_fee=None,
                        fiat_ticker=transaction[_FIAT_CURRENCY],
                        notes=notes,
                    )
                )
        return ProcessOperationResult(in_transactions=in_transaction_list, out_transactions=out_transaction_list, intra_transactions=[])

    def _process_fiat_payment(self, transaction: Any, notes: Optional[str] = None) -> ProcessOperationResult:
        self._logger.debug("Fiat Payment: %s", json.dumps(transaction))
        in_transaction_list: List[InTransaction] = []
        out_transaction_list: List[OutTransaction] = []

        if transaction[_STATUS] == "Completed":
            if self.is_native_fiat(transaction[_FIAT_CURRENCY]):

                # For double entry accounting purposes we must create a fiat InTransaction
                in_transaction_list.append(
                    InTransaction(
                        plugin=self.__PLUGIN_NAME,
                        unique_id=transaction[_ORDER_NO],
                        raw_data=json.dumps(transaction),
                        timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_CREATE_TIME]),
                        asset=transaction[_CRYPTOCURRENCY],
                        exchange=self.__EXCHANGE_NAME,
                        holder=self.account_holder,
                        transaction_type=Keyword.BUY.value,
                        spot_price=str(RP2Decimal(transaction[_PRICE])),
                        crypto_in=transaction[_OBTAIN_AMOUNT],
                        crypto_fee=None,
                        fiat_in_no_fee=str(RP2Decimal(transaction[_SOURCE_AMOUNT]) - RP2Decimal(transaction[_TOTAL_FEE])),
                        fiat_in_with_fee=str(transaction[_SOURCE_AMOUNT]),
                        fiat_fee=str(RP2Decimal(transaction[_TOTAL_FEE])),
                        fiat_ticker=transaction[_FIAT_CURRENCY],
                        notes=(f"{notes + '; ' if notes else ''}Buy transaction for native fiat payment orderNo - {transaction[_ORDER_NO]}"),
                    )
                )

                # This is an OutTransaction for a buy or conversion based on what the native fiat is.
                out_transaction_list.append(
                    OutTransaction(
                        plugin=self.__PLUGIN_NAME,
                        unique_id=transaction[_ORDER_NO],
                        raw_data=json.dumps(transaction),
                        timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_CREATE_TIME]),
                        asset=transaction[_FIAT_CURRENCY],
                        exchange=self.__EXCHANGE_NAME,
                        holder=self.account_holder,
                        transaction_type=Keyword.SELL.value,
                        spot_price=str(RP2Decimal("1")),
                        crypto_out_no_fee=str(RP2Decimal(transaction[_SOURCE_AMOUNT]) - RP2Decimal(transaction[_TOTAL_FEE])),
                        crypto_fee=str(RP2Decimal(transaction[_TOTAL_FEE])),
                        crypto_out_with_fee=str(RP2Decimal(transaction[_SOURCE_AMOUNT])),
                        fiat_out_no_fee=str(RP2Decimal(transaction[_SOURCE_AMOUNT]) - RP2Decimal(transaction[_TOTAL_FEE])),
                        fiat_fee=None,
                        fiat_ticker=transaction[_FIAT_CURRENCY],
                        notes=(f"{notes + '; ' if notes else ''}Sell transaction conversion from native fiat orderNo - " f"{transaction[_ORDER_NO]}"),
                    )
                )
            else:
                in_transaction_list.append(
                    InTransaction(
                        plugin=self.__PLUGIN_NAME,
                        unique_id=transaction[_ORDER_NO],
                        raw_data=json.dumps(transaction),
                        timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_CREATE_TIME]),
                        asset=transaction[_CRYPTOCURRENCY],
                        exchange=self.__EXCHANGE_NAME,
                        holder=self.account_holder,
                        transaction_type=Keyword.BUY.value,
                        spot_price=Keyword.UNKNOWN.value,
                        crypto_in=transaction[_OBTAIN_AMOUNT],
                        crypto_fee=None,
                        fiat_in_no_fee=None,
                        fiat_in_with_fee=None,
                        fiat_fee=None,
                        notes=(f"{notes + '; ' if notes else ''}Buy transaction conversion from non-native_fiat orderNo - {transaction[_ORDER_NO]}"),
                    )
                )

                out_transaction_list.append(
                    OutTransaction(
                        plugin=self.__PLUGIN_NAME,
                        unique_id=transaction[_ORDER_NO],
                        raw_data=json.dumps(transaction),
                        timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_CREATE_TIME]),
                        asset=transaction[_FIAT_CURRENCY],
                        exchange=self.__EXCHANGE_NAME,
                        holder=self.account_holder,
                        transaction_type=Keyword.SELL.value,
                        spot_price=Keyword.UNKNOWN.value,
                        crypto_out_no_fee=str(RP2Decimal(transaction[_SOURCE_AMOUNT]) - RP2Decimal(transaction[_TOTAL_FEE])),
                        crypto_fee=str(RP2Decimal(transaction[_TOTAL_FEE])),
                        crypto_out_with_fee=str(RP2Decimal(transaction[_SOURCE_AMOUNT])),
                        fiat_out_no_fee=None,
                        fiat_fee=None,
                        notes=(f"{notes + '; ' if notes else ''}Sell transaction conversion from non-native_fiat orderNo - " f"{transaction[_ORDER_NO]}"),
                    )
                )

            # An InTransaction is needed for the fiat in order for the accounting to zero out
            in_transaction_list.append(
                InTransaction(
                    plugin=self.__PLUGIN_NAME,
                    unique_id=f"{transaction[_ORDER_NO]}/fiat_buy",
                    raw_data=json.dumps(transaction),
                    timestamp=self._rp2_timestamp_from_ms_epoch(transaction[_CREATE_TIME]),
                    asset=transaction[_FIAT_CURRENCY],
                    exchange=self.__EXCHANGE_NAME,
                    holder=self.account_holder,
                    transaction_type=Keyword.BUY.value,
                    spot_price=str(RP2Decimal("1")),
                    crypto_in=str(RP2Decimal(transaction[_SOURCE_AMOUNT])),
                    crypto_fee=None,
                    fiat_in_no_fee=str(RP2Decimal(transaction[_SOURCE_AMOUNT])),
                    fiat_in_with_fee=str(RP2Decimal(transaction[_SOURCE_AMOUNT])),
                    fiat_fee=None,
                    fiat_ticker=transaction[_FIAT_CURRENCY],
                    notes=(f"{notes + '; ' if notes else ''}Fiat deposit for orderNo - {transaction[_ORDER_NO]}"),
                )
            )

        return ProcessOperationResult(in_transactions=in_transaction_list, out_transactions=out_transaction_list, intra_transactions=[])



# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# CCXT documentation:
# https://docs.ccxt.com/en/latest/index.html

# pylint: disable=fixme

from datetime import datetime
from typing import Any, Dict, List, NamedTuple, Optional, Union

from rp2.rp2_error import RP2RuntimeError

_TIMESTAMP: str = "timestamp"

# Time period constants
_NINETY_DAYS_IN_MS: int = 7776000000
_THIRTY_DAYS_IN_MS: int = 2592000000
_ONE_DAY_IN_MS: int = 86400000
_MS_IN_SECOND: int = 1000

_DEFAULT_WINDOW: int = _THIRTY_DAYS_IN_MS


class PaginationDetails(NamedTuple):
    symbol: Optional[str]
    since: Optional[int]
    limit: Optional[int]
    params: Optional[Dict[str, Union[int, str, None]]]


class AbstractPaginationDetailSet:
    def __iter__(self) -> "AbstractPaginationDetailsIterator":
        raise NotImplementedError("Abstract method")


class DateBasedPaginationDetailSet(AbstractPaginationDetailSet):
    def __init__(
        self,
        exchange_start_time: int,
        limit: Optional[int] = None,
        markets: Optional[List[str]] = None,
        params: Optional[Dict[str, Union[int, str, None]]] = None,
        window: Optional[int] = None,
    ) -> None:
        params = {} if params is None else params
        super().__init__()
        self.__exchange_start_time: int = exchange_start_time
        self.__limit: Optional[int] = limit
        self.__markets: Optional[List[str]] = markets
        self.__params: Optional[Dict[str, Union[int, str, None]]] = params
        self.__window: Optional[int] = window

    def __iter__(self) -> "DateBasedPaginationDetailsIterator":
        return DateBasedPaginationDetailsIterator(
            self.__exchange_start_time,
            self.__limit,
            self.__markets,
            self.__params,
            self.__window,
        )

    def _get_window(self) -> Optional[int]:
        return self.__window

    def _get_exchange_start_time(self) -> int:
        return self.__exchange_start_time

    def _get_limit(self) -> Optional[int]:
        return self.__limit

    def _get_markets(self) -> Optional[List[str]]:
        return self.__markets

    def _get_params(self) -> Optional[Dict[str, Union[int, str, None]]]:
        return self.__params


class CustomDateBasedPaginationDetailSet(DateBasedPaginationDetailSet):
    def __init__(
        self,
        exchange_start_time: int,
        start_time_key: str,
        end_time_key: str,
        window: int,
        limit: Optional[int] = None,
        markets: Optional[List[str]] = None,
        params: Optional[Dict[str, Union[int, str, None]]] = None,
    ) -> None:

        super().__init__(exchange_start_time, limit, markets, params, window)
        self.__start_time_key: str = start_time_key
        self.__end_time_key: str = end_time_key

    def _get_window(self) -> int:
        if self.__window:
            return self.__window
        raise RP2RuntimeError("No window defined for iterator.")

    def __iter__(self) -> "CustomDateBasedPaginationDetailsIterator":
        return CustomDateBasedPaginationDetailsIterator(
            self._get_exchange_start_time(),
            self.__start_time_key,
            self.__end_time_key,
            self._get_window(),
            self._get_limit(),
            self._get_markets(),
            self._get_params(),
        )


class AbstractPaginationDetailsIterator:
    def __init__(self, limit: Optional[int], markets: Optional[List[str]] = None, params: Optional[Dict[str, Union[int, str, None]]] = None) -> None:
        params = {} if params is None else params
        self.__limit: Optional[int] = limit
        self.__markets: Optional[List[str]] = markets
        self.__market_count: int = 0
        self.__params: Optional[Dict[str, Union[int, str, None]]] = params

    def _get_market(self) -> Optional[str]:
        return self.__markets[self.__market_count] if self.__markets else None

    def _has_more_markets(self) -> bool:
        return self.__market_count < (len(self.__markets) - 1) if self.__markets else False

    def _next_market(self) -> None:
        self.__market_count += 1

    def _get_limit(self) -> Optional[int]:
        return self.__limit

    def _get_params(self) -> Optional[Dict[str, Union[int, str, None]]]:
        return self.__params

    def _get_since(self) -> Optional[int]:
        return None

    def update_fetched_elements(self, current_results: Any) -> None:
        raise NotImplementedError("Abstract method")

    def __next__(self) -> PaginationDetails:
        raise NotImplementedError("Abstract method")


class DateBasedPaginationDetailsIterator(AbstractPaginationDetailsIterator):
    def __init__(
        self,
        exchange_start_time: int,
        limit: Optional[int] = None,
        markets: Optional[List[str]] = None,
        params: Optional[Dict[str, Union[int, str, None]]] = None,
        window: Optional[int] = None,
    ) -> None:

        super().__init__(limit, markets, params)
        self.__end_of_data = False
        self.__since: int = exchange_start_time
        self.__exchange_start_time: int = exchange_start_time
        self.__now: int = int(datetime.now().timestamp()) * _MS_IN_SECOND
        self.__window: Optional[int] = window

    def update_fetched_elements(self, current_results: Any) -> None:

        end_of_market: bool = False

        # Update Since if needed otherwise end_of_market
        if len(current_results) == self._get_limit():
            # All times are inclusive
            self.__since = current_results[len(current_results) - 1][_TIMESTAMP] + 1
        elif self.__window:
            self.__since += self.__window
            if self.__since > self.__now:
                end_of_market = True
        else:
            end_of_market = True

        if end_of_market:
            if self._has_more_markets():
                # we have reached the end of one market, now let's move on to the next
                self.__since = self.__exchange_start_time
                self._next_market()
            else:
                self.__end_of_data = True

    def _is_end_of_data(self) -> bool:
        return self.__end_of_data

    def _get_since(self) -> int:
        return self.__since

    def _get_end_of_window(self) -> int:
        if self.__window:
            return self.__since + self.__window
        raise RP2RuntimeError("No window defined for iterator.")

    def __next__(self) -> PaginationDetails:
        while not self._is_end_of_data():
            return PaginationDetails(
                symbol=self._get_market(),
                since=self._get_since(),
                limit=self._get_limit(),
                params=self._get_params(),
            )
        raise StopIteration(self)


class CustomDateBasedPaginationDetailsIterator(DateBasedPaginationDetailsIterator):
    def __init__(
        self,
        exchange_start_time: int,
        start_time_key: str,
        end_time_key: str,
        window: int,
        limit: Optional[int] = None,
        markets: Optional[List[str]] = None,
        params: Optional[Dict[str, Union[int, str, None]]] = None,
    ) -> None:

        super().__init__(exchange_start_time, limit, markets, params, window)
        self.__start_time_key: str = start_time_key
        self.__end_time_key: str = end_time_key

    def __next__(self) -> PaginationDetails:

        while not self._is_end_of_data():
            base_details: PaginationDetails = super().__next__()
            if base_details.params:
                base_details.params[self.__start_time_key] = base_details.since
                base_details.params[self.__end_time_key] = self._get_end_of_window()
            else:
                base_details._replace(params={self.__start_time_key: base_details.since, self.__end_time_key: self._get_end_of_window()})
            return base_details
        raise StopIteration(self)


# TODO: Add IdBasedPaginationDetails and PageNumberBasedPaginationDetails classes



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timedelta
from json import JSONDecodeError, loads
from typing import Any, Dict, List, NamedTuple, Optional, cast

import requests
from requests.exceptions import ReadTimeout
from requests.models import Response
from requests.sessions import Session
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2RuntimeError, RP2TypeError

from dali.cache import load_from_cache, save_to_cache
from dali.configuration import HISTORICAL_PRICE_KEYWORD_SET
from dali.historical_bar import HistoricalBar
from dali.logger import LOGGER

# exchangerates.host keywords
_SUCCESS: str = "success"
_SYMBOLS: str = "symbols"
_RATES: str = "rates"

# exchangerates.host urls
_EXCHANGE_BASE_URL: str = "https://api.exchangerate.host/"
_EXCHANGE_SYMBOLS_URL: str = "https://api.exchangerate.host/symbols"

_DAYS_IN_SECONDS: int = 86400
_FIAT_EXCHANGE: str = "exchangerate.host"

# First on the list has the most priority
# This is hard-coded for now based on volume of each of these markets for BTC on Coinmarketcap.com
# Any change to this priority should be documented in "docs/configuration_file.md"
_FIAT_PRIORITY: List[str] = ["USD", "JPY", "KRW", "EUR", "GBP", "AUD"]


class AssetPairAndTimestamp(NamedTuple):
    timestamp: datetime
    from_asset: str
    to_asset: str
    exchange: str


class AbstractPairConverterPlugin:
    __ISSUES_URL: str = "https://github.com/eprbell/dali-rp2/issues"
    __TIMEOUT: int = 30

    def __init__(self, historical_price_type: str, fiat_priority: Optional[str] = None) -> None:
        if not isinstance(historical_price_type, str):
            raise RP2TypeError(f"historical_price_type is not a string: {historical_price_type}")
        if historical_price_type not in HISTORICAL_PRICE_KEYWORD_SET:
            raise RP2TypeError(
                f"historical_price_type must be one of {', '.join(sorted(HISTORICAL_PRICE_KEYWORD_SET))}, instead it was: {historical_price_type}"
            )
        result = cast(Dict[AssetPairAndTimestamp, HistoricalBar], load_from_cache(self.cache_key()))
        self.__cache: Dict[AssetPairAndTimestamp, HistoricalBar] = result if result is not None else {}
        self.__historical_price_type: str = historical_price_type
        self.__session: Session = requests.Session()
        self.__fiat_list: List[str] = []
        self.__fiat_priority: List[str]
        self.__fiat_priority = loads(fiat_priority) if fiat_priority is not None else _FIAT_PRIORITY

    def name(self) -> str:
        raise NotImplementedError("Abstract method: it must be implemented in the plugin class")

    def cache_key(self) -> str:
        raise NotImplementedError("Abstract method: it must be implemented in the plugin class")

    def _add_bar_to_cache(self, key: AssetPairAndTimestamp, historical_bar: HistoricalBar) -> None:
        self.__cache[self._floor_key(key)] = historical_bar

    def _get_bar_from_cache(self, key: AssetPairAndTimestamp) -> Optional[HistoricalBar]:
        return self.__cache.get(self._floor_key(key))

    # The most granular pricing available is 1 minute, to reduce the size of cache and increase the reuse of pricing data
    def _floor_key(self, key: AssetPairAndTimestamp) -> AssetPairAndTimestamp:
        raw_timestamp: datetime = key.timestamp
        floored_timestamp: datetime = raw_timestamp - timedelta(
            minutes=raw_timestamp.minute % 1, seconds=raw_timestamp.second, microseconds=raw_timestamp.microsecond
        )
        floored_key: AssetPairAndTimestamp = AssetPairAndTimestamp(
            timestamp=floored_timestamp,
            from_asset=key.from_asset,
            to_asset=key.to_asset,
            exchange=key.exchange,
        )

        return floored_key

    @property
    def historical_price_type(self) -> str:
        return self.__historical_price_type

    @property
    def fiat_list(self) -> List[str]:
        return self.__fiat_list

    @property
    def issues_url(self) -> str:
        return self.__ISSUES_URL

    # The exchange parameter is a hint on which exchange to use for price lookups. The plugin is free to use it or ignore it.
    def get_historic_bar_from_native_source(self, timestamp: datetime, from_asset: str, to_asset: str, exchange: str) -> Optional[HistoricalBar]:
        raise NotImplementedError("Abstract method: it must be implemented in the plugin class")

    def save_historical_price_cache(self) -> None:
        save_to_cache(self.cache_key(), self.__cache)

    def get_conversion_rate(self, timestamp: datetime, from_asset: str, to_asset: str, exchange: str) -> Optional[RP2Decimal]:
        result: Optional[RP2Decimal] = None
        historical_bar: Optional[HistoricalBar] = None
        key: AssetPairAndTimestamp = AssetPairAndTimestamp(timestamp, from_asset, to_asset, exchange)
        log_message_qualifier: str = ""
        if key in self.__cache:
            historical_bar = self.__cache[key]
            log_message_qualifier = "cache of "
        else:
            historical_bar = self.get_historic_bar_from_native_source(timestamp, from_asset, to_asset, exchange)
            if historical_bar:
                self.__cache[key] = historical_bar

        if historical_bar:
            result = historical_bar.derive_transaction_price(timestamp, self.__historical_price_type)
            LOGGER.debug(
                "Fetched %s conversion rate %s for %s/%s->%s from %splugin %s: %s",
                self.__historical_price_type,
                result,
                timestamp,
                from_asset,
                to_asset,
                log_message_qualifier,
                self.name(),
                historical_bar,
            )

        return result

    def _build_fiat_list(self) -> None:
        try:
            response: Response = self.__session.get(_EXCHANGE_SYMBOLS_URL, timeout=self.__TIMEOUT)
            # {
            #     'motd':
            #         {
            #             'msg': 'If you or your company ...',
            #             'url': 'https://exchangerate.host/#/donate'
            #         },
            #     'success': True,
            #     'symbols':
            #         {
            #             'AED':
            #                 {
            #                     'description': 'United Arab Emirates Dirham',
            #                     'code': 'AED'
            #                 },
            #             ...
            #         }
            # }
            data: Any = response.json()
            if data[_SUCCESS]:
                self.__fiat_list = [fiat_iso for fiat_iso in data[_SYMBOLS] if fiat_iso != "BTC"]
            else:
                if "message" in data:
                    LOGGER.error("Error %d: %s: %s", response.status_code, _EXCHANGE_SYMBOLS_URL, data["message"])
                response.raise_for_status()

        except JSONDecodeError as exc:
            LOGGER.info("Fetching of fiat symbols failed. The server might be down. Please try again later.")
            raise RP2RuntimeError("JSON decode error") from exc

    def _add_fiat_edges_to_graph(self, graph: Dict[str, Dict[str, None]], markets: Dict[str, List[str]]) -> None:
        if not self.__fiat_list:
            self._build_fiat_list()

        for fiat in self.__fiat_list:
            to_fiat_list: Dict[str, None] = dict.fromkeys(self.__fiat_list.copy())
            del to_fiat_list[fiat]
            if graph.get(fiat):
                for to_be_added_fiat in to_fiat_list:
                    # add a pair if it doesn't exist
                    if to_be_added_fiat not in graph[fiat]:
                        graph[fiat][to_be_added_fiat] = None
            else:
                graph[fiat] = to_fiat_list

            for to_fiat in to_fiat_list:
                fiat_market = f"{fiat}{to_fiat}"
                markets[fiat_market] = [_FIAT_EXCHANGE]

            # Add prioritized fiat at the beginning
            for priority_fiat in reversed(self.__fiat_priority):
                if priority_fiat in graph[fiat]:
                    graph[fiat].pop(priority_fiat)
                    remainder: Dict[str, None] = graph[fiat]
                    graph[fiat] = {priority_fiat: None}
                    graph[fiat].update(remainder)

            LOGGER.debug("Added to assets for %s: %s", fiat, graph[fiat])

    def _is_fiat_pair(self, from_asset: str, to_asset: str) -> bool:
        return self._is_fiat(from_asset) and self._is_fiat(to_asset)

    def _is_fiat(self, asset: str) -> bool:
        if not self.__fiat_list:
            self._build_fiat_list()

        return asset in self.__fiat_list

    def _get_fiat_exchange_rate(self, timestamp: datetime, from_asset: str, to_asset: str) -> Optional[HistoricalBar]:
        result: Optional[HistoricalBar] = None
        params: Dict[str, Any] = {"base": from_asset, "symbols": to_asset}
        request_count: int = 0
        # exchangerate.host only gives us daily accuracy, which should be suitable for tax reporting
        while request_count < 5:
            try:
                response: Response = self.__session.get(f"{_EXCHANGE_BASE_URL}{timestamp.strftime('%Y-%m-%d')}", params=params, timeout=self.__TIMEOUT)
                # {
                #     'motd':
                #         {
                #             'msg': 'If you or your company ...',
                #             'url': 'https://exchangerate.host/#/donate'
                #         },
                #     'success': True,
                #     'historical': True,
                #     'base': 'EUR',
                #     'date': '2020-04-04',
                #     'rates':
                #         {
                #             'USD': 1.0847, ... // float, Lists all supported currencies unless you specify
                #         }
                # }
                data: Any = response.json()
                if data[_SUCCESS]:
                    result = HistoricalBar(
                        duration=timedelta(seconds=_DAYS_IN_SECONDS),
                        timestamp=timestamp,
                        open=RP2Decimal(str(data[_RATES][to_asset])),
                        high=RP2Decimal(str(data[_RATES][to_asset])),
                        low=RP2Decimal(str(data[_RATES][to_asset])),
                        close=RP2Decimal(str(data[_RATES][to_asset])),
                        volume=ZERO,
                    )
                break

            except (JSONDecodeError, ReadTimeout) as exc:
                LOGGER.debug("Fetching of fiat exchange rates failed. The server might be down. Retrying the connection.")
                request_count += 1
                if request_count > 4:
                    LOGGER.info("Giving up after 4 tries. Saving to Cache.")
                    self.save_historical_price_cache()
                    raise RP2RuntimeError("JSON decode error") from exc

        return result
